/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.ciphering;

import com.monica.javacom.ciphering.*;
import com.monica.javacom.util.*;

/**	Implements an AES (American Encription Standard, also known as RC6) encoder/decoder, which is the recent standard for symmetric-key-ciphering of digital data.
		This object uses an AESKey as the encoding/decoding key, generated by an AESKeyGenerator.
		As the main functionality, the coder gets a String which contains the initial data, and returns the coded representation of this String.
		@see AESKey
		@see AESKeyGenerator
*/

public class AES {

	private static final int p=0xB7E15163, q=0x9E3779B9;

	private AESKey key;
	int numberOfRounds;
  private MJCConverter converter;

	int[] l, s;

/**	Constructs a new AES encoder/decoder object.
		@param in_key the AESKey that the AES will use to encode/decode
		@param in_numberOfRounds the number of runds defined in the AES algorithm - default: 20
		@throws BadInputValueException the in_numberOfRounds parameter must be between 1 and 256
*/

	public AES(AESKey in_key, int in_numberOfRounds) throws BadInputValueException {

		key=in_key;
		if ((in_numberOfRounds<1) || (in_numberOfRounds>256)) throw new BadInputValueException("AES: numberOfRounds must be between 1 and 256");
		else numberOfRounds=in_numberOfRounds;

    converter=new MJCConverter();

		l=key.getKey(); int c=l.length;
		s=new int[2*numberOfRounds+4];
		s[0]=p;
		for(int k=1; k<2*numberOfRounds+4; k++)
			s[k]=s[k-1]+q;
		int a, b, i, j; a=b=i=j=0;
		int v=3*Math.max(c, 2*numberOfRounds+4);
		for(int k=1; k<=v; k++) {
			a=s[i]=rotateLeft(s[i]+a+b, 3);
			b=l[j]=rotateLeft(l[j]+a+b, a+b);
			i=(i+1)%(2*numberOfRounds+4);
			j=(j+1)%c;
		}

	}

/**	Encodes the specified String.
		@param text the input text which has to be encoded
		@return the encoded representation of the input text
		@throws ZeroLengthStringException input string must not be empty
*/

	public String encode(String text) throws ZeroLengthStringException {

		if (text.length()==0) throw new ZeroLengthStringException("AES: encode: input String must not be empty");

		char hasznosByteok; String byteokStr;
		int actual=0, maxPos=text.length();
		StringLista blocks=new StringLista();
		byte[] byteok; byte[] codedbyteok=new byte[16];
		char[] winFix=new char[16];
		String coded=new String();

		while (maxPos-actual > 16)
			blocks.add(text.substring(actual, actual+=16));
		if (maxPos-actual>0)
			blocks.add( text.substring(actual) );

		int[] intek=new int[16]; int a, b, c, d; int reg; int t, u;
		while (!blocks.isEmpty()) {
		   byteokStr=blocks.getFirst();
		   hasznosByteok=(char)byteokStr.length();
			if (hasznosByteok<16)
				byteokStr=to16Bytes(byteokStr);
                        byteok=converter.getBytes(byteokStr);

		   for(int i=0; i<16; i++)
				intek[i]=byteok[i]&0x000000FF;
		   a=(intek[0]) | (intek[1]<<8) | (intek[2]<<16) | (intek[3]<<24);
		   b=(intek[4]) | (intek[5]<<8) | (intek[6]<<16) | (intek[7]<<24);
		   c=(intek[8]) | (intek[9]<<8) | (intek[10]<<16) | (intek[11]<<24);
		   d=(intek[12]) | (intek[13]<<8) | (intek[14]<<16) | (intek[15]<<24);

			b=b+s[0];
		   d=d+s[1];
		   for(int i=1; i<=numberOfRounds; i++) {
		 		t=rotateLeft(b*(2*b+1), 5);
		 		u=rotateLeft(d*(2*d+1), 5);
		 		a=rotateLeft(a^t, u)+s[2*i];
		 		c=rotateLeft(c^u, t)+s[2*i+1];
		 		reg=a; a=b; b=c; c=d; d=reg;
		   }
		   a=a+s[2*numberOfRounds+2];
		   c=c+s[2*numberOfRounds+3];

		   codedbyteok[0]=(byte)((a&0x000000ff));
		   codedbyteok[1]=(byte)((a&0x0000ff00)>>8);
		   codedbyteok[2]=(byte)((a&0x00ff0000)>>16);
		   codedbyteok[3]=(byte)((a&0xff000000)>>24);
		   codedbyteok[4]=(byte)((b&0x000000ff));
		   codedbyteok[5]=(byte)((b&0x0000ff00)>>8);
		   codedbyteok[6]=(byte)((b&0x00ff0000)>>16);
		   codedbyteok[7]=(byte)((b&0xff000000)>>24);
		   codedbyteok[8]=(byte)((c&0x000000ff));
		   codedbyteok[9]=(byte)((c&0x0000ff00)>>8);
		   codedbyteok[10]=(byte)((c&0x00ff0000)>>16);
		   codedbyteok[11]=(byte)((c&0xff000000)>>24);
		   codedbyteok[12]=(byte)((d&0x000000ff));
		   codedbyteok[13]=(byte)((d&0x0000ff00)>>8);
		   codedbyteok[14]=(byte)((d&0x00ff0000)>>16);
		   codedbyteok[15]=(byte)((d&0xff000000)>>24);

      coded=coded+hasznosByteok;
      coded=coded.concat(converter.toString(codedbyteok));
		}

		return coded;

	}

/**	Decodes the specified String.
		@param text the encoded text which must be the result of the encode() function of an AES object whith the same key, that this AES is working with.
		@return the decoded representation of the encoded text, the input text of the encoding algorithm
		@throws ZeroLengthStringException input string must not be empty and the length of it must be a multiple of 17
*/

	public String decode(String text) throws ZeroLengthStringException {

		if (text.length()==0) throw new ZeroLengthStringException("AES: decode: input String must not be empty");
		if (text.length()%17!=0) throw new ZeroLengthStringException("AES: decode: input was not multiple of 17");

		char hasznosByteok; String byteokStr;
		int actual=0, maxPos=text.length();
		StringLista blocks=new StringLista();
		byte[] byteok; byte[] codedbyteok=new byte[16];
		char[] winFix=new char[16];
		String coded=new String();

                byteok=converter.getBytes(text);
		while (maxPos-actual >= 17) {
			blocks.add(text.substring(actual, actual+=17));
		}

		int[] intek=new int[16]; int a, b, c, d; int reg; int t, u;
		while (!blocks.isEmpty()) {
		   byteokStr=blocks.getFirst();
		   hasznosByteok=byteokStr.charAt(0);
			byteokStr=byteokStr.substring(1);
                        byteok=converter.getBytes(byteokStr);

		   for(int i=0; i<16; i++)
            intek[i]=byteok[i]&0x000000FF;
		   a=(intek[0]) | (intek[1]<<8) | (intek[2]<<16) | (intek[3]<<24);
		   b=(intek[4]) | (intek[5]<<8) | (intek[6]<<16) | (intek[7]<<24);
		   c=(intek[8]) | (intek[9]<<8) | (intek[10]<<16) | (intek[11]<<24);
		   d=(intek[12]) | (intek[13]<<8) | (intek[14]<<16) | (intek[15]<<24);

		   c=c-s[2*numberOfRounds+3];
		   a=a-s[2*numberOfRounds+2];
		   for(int i=numberOfRounds; i>=1; i--) {
		 		reg=d; d=c; c=b; b=a; a=reg;
		 		u=rotateLeft(d*(2*d+1), 5);
		 		t=rotateLeft(b*(2*b+1), 5);
		 		c=rotateRight(c-s[2*i+1], t)^u;
		 		a=rotateRight(a-s[2*i], u)^t;
		   }
		   d=d-s[1];
		   b=b-s[0];

		   codedbyteok[0]=(byte)((a&0x000000ff));
		   codedbyteok[1]=(byte)((a&0x0000ff00)>>8);
		   codedbyteok[2]=(byte)((a&0x00ff0000)>>16);
		   codedbyteok[3]=(byte)((a&0xff000000)>>24);
		   codedbyteok[4]=(byte)((b&0x000000ff));
		   codedbyteok[5]=(byte)((b&0x0000ff00)>>8);
		   codedbyteok[6]=(byte)((b&0x00ff0000)>>16);
		   codedbyteok[7]=(byte)((b&0xff000000)>>24);
		   codedbyteok[8]=(byte)((c&0x000000ff));
		   codedbyteok[9]=(byte)((c&0x0000ff00)>>8);
		   codedbyteok[10]=(byte)((c&0x00ff0000)>>16);
		   codedbyteok[11]=(byte)((c&0xff000000)>>24);
		   codedbyteok[12]=(byte)((d&0x000000ff));
		   codedbyteok[13]=(byte)((d&0x0000ff00)>>8);
		   codedbyteok[14]=(byte)((d&0x00ff0000)>>16);
		   codedbyteok[15]=(byte)((d&0xff000000)>>24);

      coded=coded.concat((converter.toString(codedbyteok)).substring(0, hasznosByteok));
		}

		return coded;

	}

/**	Extends the supplied String to a multiple length of 16. The trailing spaces are filled with (char) 0.
		@param in_text the text, which has to be extended to a multiple of 16
		@return a String whose first in_text.length() characters are the same as in in_text, but it has some trailing (char) 0 -s on the end, and whose length is a multiple of 16.
*/	
	
	private String to16Bytes(String in_text) {
						
		int len=in_text.length();
		String text=in_text;
		char nulla=0;

		for(int i=len; i<16; i++)
			text=text+nulla;
			
		return text;
	
	}
	
/**	Primitive operation missing from the standard java.math operations. Rotating to the left.
		That means shifting the values in an integer register to the left, and filling the shifted-out bits in on the LSB side.
		@param mit the integer value that has to be rotated
		@param mennyivel this value represents, with how many bits the above value has to be rotated
		@return the value of the rotated register
*/	
	
	private int rotateLeft(int mit, int mennyivel) {
	
		mennyivel=mennyivel&0x1F;
	
		for(int i=0; i<mennyivel; i++) {
		 	if(mit>0)
		 		mit=mit<<1;
		 	else
		 		mit=(mit<<1)|0x00000001;
		}
		
		return mit;
	
	}
	
/**	Primitive operation missing from the standard java.math operations. Rotating to the right.
		That means shifting the values in an integer register to the right, and filling the shifted-out bits in on the MSB side.
		@param mit the integer value that has to be rotated
		@param mennyivel this value represents, with how many bits the above value has to be rotated
		@return the value of the rotated register
*/	
	private int rotateRight(int mit, int mennyivel) {
	
		mennyivel=mennyivel&0x1F;
		
		for(int i=0; i<mennyivel; i++) {
		 	if((mit&0x01)==0)
		 		mit=mit>>>1;
		 	else
		 		mit=(mit>>1)|0x80000000;
		}
		
		return mit;
	
	}
	
/**	Returns the AESKey of this AES object.
		@return the AESKey, that this object is working with
*/	
	
	public AESKey getKey() {
	
		return key;
	
	}
	
/**	Returns the number of rounds of this AES object.
		@return the value, that this object is working with as number of runds definded in the AES algorithm.
*/	

	public int getNumberOfRounds() {
	
		return numberOfRounds;
	
	}

/**	Sets the AESKey of this object.
		@param in_key the new AESKey that this AES object should use
*/

   public void setKey(AESKey in_key) {

   	key=in_key;

   }

/**	Sets the number of rounds of this object.
		@param in_numberOfRounds the number of runds defined in the AES algorithm
		@throws BadInputValueException the in_numberOfRounds parameter must be between 1 and 256
*/

   public void setNumberOfRounds(int in_numberOfRounds) throws BadInputValueException {

		if ((in_numberOfRounds<1) || (in_numberOfRounds>256)) throw new BadInputValueException("AES: numberOfRounds must be between 1 and 256");
		else numberOfRounds=in_numberOfRounds;

   }

}
