/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.ciphering;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;

import java.math.BigInteger;
import java.util.Random;

/**	This is a middle level layer in the MJC layer hierarchy responsible for ciphering the data stream between the upper layers.
		Implements cipher key exchanges, coding and encoding.
		At initialization it is transparent, it leaves data as it is, it just lets input through.
		After getting setTransparent(false) the keyexchange starts. The client side has the setTransparent(false) called by the upper layer.
		It sends a special message to the server, that also calls its own setTransparent(false). The server generates an rsa-key-pair, sends
		the public key to the client in a special packet. After that, he constructs his own RSA coder with the private key of the keypair just generated.
		The client, after receiving the public key, constructs his RSA coder, and generates an aes key. He sends this key, coded with the public key, to
		the server, and constructs his AES coder.
		The server, as he got the AES key, also constructs the AES coder. So, by now, both sides have the proper AES coder (hopefully). The server sends
		a special String, included in a control pocket, to the client, and switches to ciphering mode. From this point he can only recieve encripted data,
		this means, every pocket, he gets will be decripted before forwarding it to the upper layer.
		The client, as it gets the special control packet, tries to decript it. If the original String is the result, then everything worked fine during
		the key exchange, and it also switches to ciphering mode, like the server. At this point the setTransparent(false) function returns from its
		synchronized state, and the communication (telnet initialization, authentification, data exchange) can go on encripted.
*/

public class MJCBasicCiphering implements MJCCiphering {

  private static final String NAME="MJCBasicCiphering";
  private static final String VERSION="0.1";

        private final static int rsakeyLength = 32;
	private final static int certainty = 100;
	private final static int publicKey = 65537;
	private final static int aeskeyLength = 16;

  private MJCReadable readable;
  private MJCReader reader;
  private MJCWriteable writeable;
  private MJCConfiguration configuration;

  private boolean threadSuspended;
  private boolean threadSuspendedForReadout;
  private boolean threadRunning;
  private boolean waitForKeyExchange;

  private boolean isTransparent;

  private boolean enVagyokASzerver;
  private RSA rsaKodolo;
  private RSAKey rsaKulcs;
  private AES aesKodolo;
  private AESKey aesKulcs;
  private String normal="n", control="c";

  private MJCStringBuffer readBuffer;

/**	Constructs a new MJCCiphering layer.
		@param in_enVagyokASzerver true if this is the server side, false if this is the client side
*/

  public MJCBasicCiphering(boolean in_enVagyokASzerver) {
    Thread readerThread=new Thread(this);
    threadRunning=true;
    threadSuspended=true;
    threadSuspendedForReadout=true;
    readBuffer=new MJCStringBuffer(50000);
    readerThread.start();

    enVagyokASzerver=in_enVagyokASzerver;
    if (enVagyokASzerver) cipherOutput("Letrejottem, en vagyok a szerver."); else cipherOutput("Letrejottem, en vagyok a kliens.");

    isTransparent=true;
    cipherOutput("isTransparent="+isTransparent);
    waitForKeyExchange=false;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCCiphering */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCCiphering.
      @return the MJC name of the MJCCiphering
  */

  public String getCipheringName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCCiphering.
      @return the MJC version of the MJCCiphering
  */

  public String getCipheringVersion() {
    return VERSION;
  }

  /** This method toggles the transparency mode of the MJCCiphering layer.
      The parameter true makes the layer transparent, this means no
      encoding will happen, the parameter false shows the beginning of
      the encoding. By default the layer is transparent. Currently only
      setTransparent(false) is supported, setTransparent(true) has no effect.
      @param status the desired transparency mode
  */

  public synchronized void setTransparent(boolean status) {
		cipherOutput("Meghivtak a setTransparent-et.");
		if (!status) {
			if (enVagyokASzerver) {
				cipherOutput("...megpedig a szerveret.");
				try {
					RSAKeyPairGenerator rsagenerator=new RSAKeyPairGenerator(rsakeyLength, certainty, publicKey, new Random());
					RSAKeyPair rsapair=rsagenerator.generateRSAKeyPair();
					rsaKulcs=rsapair.getPrivate();
					RSAKey kliensKulcsa=rsapair.getPublic();
					cipherOutput("Szerver: kulcsok generalva.");
					cipherOutput("Szerver: titkos kulcs: "+rsaKulcs.toString());
					cipherOutput("Szerver: nyilvanos kulcs:"+kliensKulcsa.toString());
					String klstr=kliensKulcsa.toString();
					writeable.write("p"+(char)(klstr.length()/256)+(char)(klstr.length()%256)+klstr); flush();
					cipherOutput("Szerver: atkuldtem a nyilvanos kulcsot.");
					rsaKodolo=new RSA(rsaKulcs);
					cipherOutput("Szerver: Letrehoztam az RSA kodolot.");
				}
				catch (BadInputValueException e) {
					cipherOutput("Szerver: BadInputValueException (RSA)");
				}
			}
			else {
				cipherOutput("...megpedig a klienset.");
				writeable.write("c"); flush();
				waitForKeyExchange=true;
				synchronized (this) {
					while (waitForKeyExchange) {
						try {
							wait();
						}
						catch (Exception e) {
							cipherOutput("setTransparent(): wait exception");
						}
					}
				}
			}
		}
		else
			cipherOutput("Gaz van, a setTransparent(boolean)-t valaki true-val hivta");
  }

/**	This is an important function for the keyexhange, though it is private.
*/
	
	private void keyExchange(int lepes, String adat) {
		switch (lepes) {
			case 1:
				cipherOutput("keyExchange: 1 lepes");
				try {
					rsaKulcs=new RSAKey(adat);
				 	cipherOutput("A megkapott publicKey a kovetkezo: "+rsaKulcs.toString());
				}
				catch (LoadKeyException e) {
					cipherOutput("GAZ : A megkapott publicKey rossz.");
				}
				rsaKodolo=new RSA(rsaKulcs);
				cipherOutput("Kliens: Letrehoztam az RSA kodolot.");
				try {
					AESKeyGenerator aesgenerator=new AESKeyGenerator(aeskeyLength, new Random());
					aesKulcs=aesgenerator.generateAESKey();
					cipherOutput("Kliens: AES kulcs generalva.");
					cipherOutput("Kliens: AES kulcs: "+aesKulcs.toString());
					String sestr=aesKulcs.toString();
					try {
						sestr=rsaKodolo.encode(sestr);
					}
					catch (ZeroLengthStringException e) {
						cipherOutput("Az elso lepesben az RSA kodolo 0 hosszu Stringet kapott.");
					}

					char ccc[]=new char[2];
					ccc[0]=(char)(sestr.length()/256);
					ccc[1]=(char)(sestr.length()%256);
					String sss="a"+new String(ccc)+sestr;

					writeable.write(sss); flush();
					cipherOutput("Kliens: atkuldtem az AES kulcsot.");
					aesKodolo=new AES(aesKulcs, 20);
					cipherOutput("Kliens: Letrehoztam az AES kodolot.");
				}
				catch (BadInputValueException e) {
					cipherOutput("Kliens: BadInputValueException (AES)");
				}
				break;
			case 2:
				cipherOutput("keyExchange: 2 lepes");
				try {
					try {
						cipherOutput("Dekodolas elott az aesKulcs: "+adat);
						adat=rsaKodolo.decode(adat);
						cipherOutput("Dekodolas utan az aesKulcs: "+adat);
					}
					catch (ZeroLengthStringException e) {
						cipherOutput("A masodik lepesben az RSA dekodolo 0 hosszu Stringet kapott.");
					}
					aesKulcs=new AESKey(adat);
				 	cipherOutput("A megkapott AES kulcs a kovetkezo: "+aesKulcs.toString());
				}
				catch (LoadKeyException e) {
					cipherOutput("GAZ : A megkapott AES Kulcs rossz.");
				}
				try {
					aesKodolo=new AES(aesKulcs, 20);
					cipherOutput("Szerver: Letrehoztam az AES kodolot.");
				}
				catch (BadInputValueException e) {
					cipherOutput("Szerver: baj tortent az AES letrehozasakor.");
				}
				try {
					String okString=aesKodolo.encode("<Kulcscsere OK!>");
					writeable.write("o"+(char)(okString.length()/256)+(char)(okString.length()%256)+okString); flush();
					cipherOutput("Megirtam, hogy <Kulcscsere OK!>");
					isTransparent=false;
					cipherOutput("Innentol kodolva folytatjuk.");
					cipherOutput("isTransparent="+isTransparent);
				}
				catch (ZeroLengthStringException e) {
					cipherOutput("A masodik lepesben nem tudtam atkuldeni a nyugtazo szoveget.");
				}
				break;
			default:
				cipherOutput("GAZ : A keyExchange nem deffinialt lepessel (int) lett meghivva.");
				break;
		}
	}

  /* ---------------------------------------------------------------------------------- */
  /* MJCWriter */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCWriter.
      @return the MJC name of the MJCWriter
  */

  public String getWriterName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCWriter.
      @return the MJC version of the MJCWriter
  */

  public String getWriterVersion() {
    return VERSION;
  }

  /** Sets the MJCWriteable to which this MJCWriter will write.
      @param writeable the MJCWriteable to which this MJCWriter will write
  */	

  public void setWriteable(MJCWriteable writeable) {
    this.writeable=writeable;
  }

  /** Returns the MJCWriteable to which this MCJWriter will write.
      @return the MJCWriteable to which this MJCWriter will write
  */	

  public MJCWriteable getWriteable() {
    return writeable;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCConfigureable */
  /* ---------------------------------------------------------------------------------- */

  /** Sets the MJCConfiguration for this object.
      @param configuration the MJCConfiguration object for the customization
  */	

  public void setConfiguration(MJCConfiguration configuration) {
    this.configuration=configuration;
  }

  /** Returns the MJCConfiguration with which this object is customized.
      @return the MJCConfiguration with which this object is customized
  */	

  public MJCConfiguration getConfiguration() {
    return configuration;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCReadable */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCReadable.
      @return the MJC name of the MJCReadable
  */

  public String getReadableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReadable.
      @return the MJC version of the MJCReadable
  */

  public String getReadableVersion() {
    return VERSION;
  }

  /** Returns wether the MJCReadable is still connected to its source.
      @return wether the MJCReadable is still connected to its source
  */	

  public boolean isReadableConnected() {
    if (readable!=null) {
      return readable.isReadableConnected();
    }
    else {
      return false;
    }
  }

  /** Returns the number of bytes that are available for read out from this MJCReadable.
      @return the number of bytes available for read out
  */

  public int getAvailable() {
    return readBuffer.getBufferLength();
  }

  /** Returns the contents of the read buffer from this MJCReadable.
      @return the contents of the read buffer
  */

  public synchronized String getInput() {
    String temp=readBuffer.emptyBuffer();
    threadSuspendedForReadout=false;
    notify();
    return temp;
  }

  /** Sets the MJCReader to which the readEvents have to be sent.
      @param reader the MJCReader to which the readEvents have to be sent
  */	

  public void setReader(MJCReader reader) {
    this.reader=reader;
    if (readBuffer.getBufferLength()>0) {
      reader.readEvent(this);
    }
  }

  /** Returns the MJCReader to which readEvents will be sent.
      @return the MJCReader to which readEvents will be sent
  */

  public MJCReader getReader() {
    return reader;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCReader */
  /* ---------------------------------------------------------------------------------- */

  /** Sets the MJCReadable from which data has to be read.
      @param readable the MJCReadable from which data has to be read
  */	

  public void setReadable(MJCReadable readable) {
    this.readable=readable;
  }

  /** Returns the MJCReadable from which data will be read.
      @return the MJCReadable from which data will be read
  */

  public MJCReadable getReadable() {
    return readable;
  }

/**	Run thread. The function of it is reading the data from the lower layer, and transferring it to the readbuffer.
*/

	public void run() {

		String temp;
		String buffer= new String();
		int len;
		
		while (threadRunning) {
			synchronized(this) {
				while (threadSuspended) {
					try {
						wait();
					}
					catch (Exception e) {
						cipherOutput("run(): wait exception");
					}
				}
				threadSuspended=true;
			}
			temp=readable.getInput();
//			if (temp.length()>=4) System.out.println("e:"+(int)temp.charAt(3));

			while (temp.length()>0) {  			 //temp-et fogjuk feldolgozni: addig csinaljuk, amig meg van belole hatra valahany karakter
				if(isTransparent) {
					cipherOutput("Beleptem a kodolas nelkuli agba.");
					switch (temp.charAt(0)) {
						case 'n':
							cipherOutput("'n' volt a soron kovetkezo karakter");
							if(temp.length()>3) {
								len=((int)temp.charAt(1))*256+((int)temp.charAt(2));
								readBuffer.addToBuffer(temp.substring(3,3+len));
								temp=temp.substring(3+len);
							}
							else {
								cipherOutput("GAZ : 'n' volt az elso, de kevesebb, mint 3 karakter volt hatra.");
							}
							break;
						case 'c':
							cipherOutput("'c' volt a soron kovetkezo karakter");
							setTransparent(false);
							temp=temp.substring(1);
							break;
						case 'p':
							cipherOutput("'p' volt a soron kovetkezo karakter");
							if(temp.length()>3) {
								len=((int)temp.charAt(1))*256+((int)temp.charAt(2));
								String tovabbadni=temp.substring(3,3+len);
								temp=temp.substring(3+len);
								keyExchange(1, tovabbadni);
							}
							else {
								cipherOutput("GAZ : 'p' volt az elso, de kevesebb, mint 3 karakter volt hatra.");
							}
							break;
						case 'a':
							cipherOutput("'a' volt a soron kovetkezo karakter");
							if(temp.length()>3) {
								len=((int)temp.charAt(1))*256+((int)temp.charAt(2));
								String tovabbadni=temp.substring(3,3+len);
								temp=temp.substring(3+len);
								keyExchange(2, tovabbadni);
							}
							else {
								cipherOutput("GAZ : 'a' volt az elso, de kevesebb, mint 3 karakter volt hatra.");
							}
							break;
						case 'o':
							cipherOutput("'o' volt a soron kovetkezo karakter");
							if(temp.length()>3) {
								len=((int)temp.charAt(1))*256+((int)temp.charAt(2));
								try {
									if ((aesKodolo.decode(temp.substring(3,3+len))).equals("<Kulcscsere OK!>")) {
										cipherOutput("Megkaptam, hogy <Kulcscsere OK!>");
										isTransparent=false;
										cipherOutput("Innentol kodolva folytatjuk.");
										cipherOutput("isTransparent="+isTransparent);
										waitForKeyExchange=false;
										synchronized (this) {
											try {
												notify();
											}
											catch (Exception e) {
												cipherOutput("run(): notify exception");
											}
										}
									}
									else {
										cipherOutput("Ez az amit nem kellett volna kapni:"+aesKodolo.decode(temp.substring(3,3+len)));
										cipherOutput("Kaptam valamit, de nem azt, hogy <Kulcscsere OK!>");
									}
								}
								catch (ZeroLengthStringException e) {
									cipherOutput("Baj tortent az 'o' ciklusban: a megkapott String 0 hosszusagu volt.");
								}
								temp=temp.substring(3+len);
							}
							else {
								cipherOutput("GAZ : 'o' volt az elso, de kevesebb, mint 3 karakter volt hatra.");
							}
							break;
						default:
							cipherOutput("GAZ : A kovetkezo betu nemdeffinialt volt.");
							temp=temp.substring(1);
							break;
					}

					if (reader!=null && readBuffer.getBufferLength()>0) {
						reader.readEvent(this);
					}
				}
				else {
					buffer=buffer.concat(temp);
					temp=new String();
					if (buffer.length()%17==0) {
						cipherOutput("A kapott String volt 17-tel oszthato volt, ezert most kodolok.");
						try {
		  					if (readBuffer.addToBuffer(aesKodolo.decode(buffer))) {
				  				if (reader!=null && readBuffer.getBufferLength()>0) {
					  				reader.readEvent(this);
            	         	buffer=new String();
								}
							}
						}
						catch (ZeroLengthStringException e) {
							cipherOutput("Az aesKodolo az olvasas (run()) agban 0 hosszu Stringet kapott.");
						}
					}
					else {
						cipherOutput("A kapott String nem volt 17-tel oszthato, ezert meg varok.");
					}
				}
			}
		}
		cipherOutput("Read Thread Stopped.");
	}

  /** Returns the MJC name of the MJCReader.
      @return the MJC name of the MJCReader
  */

  public String getReaderName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReader.
      @return the MJC version of the MJCReader
  */

  public String getReaderVersion() {
    return VERSION;
  }

  /** Indicates that there is data available from the specified MJCReadable.
      @param readable there is data available from the readable
  */	

  public synchronized void readEvent(MJCReadable readable) {
    threadSuspended=false;
    notifyAll();
//    cipherOutput("readEvent()");
  }

  /** Indicates that the source connection of the specified MJCReadable has been closed.
      @param readable the MJCReadable whose source connection has been closed
  */	

  public synchronized void connectionClosedEvent(MJCReadable readable) {
    threadRunning=false;
    threadSuspended=false;
    try {
      notify();
    }
    catch (Exception e) {
		 cipherOutput("connectionClosedEvent(): notify exception");
    }

    if (reader!=null) {
      reader.connectionClosedEvent(this);
    }
  }

  /** Indicates that there was an error in the source connection of the specified MJCReadable.
      @param errorDescription the error message
      @param readable the MJCReadable where the error occured
  */	

  public void errorEvent(String errorDescription, MJCReadable readable) {
    if (reader!=null) {
      reader.errorEvent(errorDescription, this);
    }
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCWriteable */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCWriteable.
      @return the MJC name of the MJCWriteable
  */

  public String getWriteableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCWriteable.
      @return the MJC version of the MJCWriteable
  */

  public String getWriteableVersion() {
    return VERSION;
  }

  /** Writes one single char. Writing is synchronous and can be buffered.
      @param c the char to be written
      @return wether the writing has been successful
  */
	
	public boolean write(char c) {
//		cipherOutput("write()");
		if (writeable!=null) {
			if (isTransparent) {
				return writeable.write(normal+((char)0)+((char)1)+c);
			}
			else {
				char[] ct=new char[1]; ct[0]=c;
				try {
  					return writeable.write(aesKodolo.encode(new String(ct)));
				}
				catch (ZeroLengthStringException e) {
					cipherOutput("Az aesKodolo az iras (write()) agban 0 hosszu Stringet kapott.");
					return false;
				}
			}
		}
		else {
			return false;
		}
	}

  /** Writes a char array. Writing is synchronous and can be buffered.
      @param c the char array to be written
      @return wether the writing has been successful
  */
	
	public boolean write(char[] c) {
//		cipherOutput("write()");
		if (writeable!=null) {
			if (isTransparent) {
				if (c.length>0)
					return writeable.write(normal+(char)(c.length/256)+(char)(c.length%256)+String.valueOf(c));
				else
					return true;
			}
			else {
				try {
  					return writeable.write(aesKodolo.encode(new String(c)));
				}
				catch (ZeroLengthStringException e) {
					cipherOutput("Az aesKodolo az iras (write()) agban 0 hosszu Stringet kapott.");
					return false;
				}
			}
		}
		else {
			return false;
		}
	}

  /** Writes a String. Writing is synchronous and can be buffered.
      @param s the String to be written
      @return wether the writing has been successful
  */
	
	public boolean write(String s) {
//		cipherOutput("write()");
		if (writeable!=null) {
			if (isTransparent) {
				if (s.length()>0)
					return writeable.write(normal+(char)(s.length()/256)+(char)(s.length()%256)+s);
				else
					return true;
			}
			else {
				try {
  					return writeable.write(aesKodolo.encode(s));
				}
				catch (ZeroLengthStringException e) {
					cipherOutput("Az aesKodolo az iras (write()) agban 0 hosszu Stringet kapott.");
					return false;
				}
			}
		}
		else {
			return false;
		}
	}

  /** Flushes the contents of the write buffer to bottom layers.
  */

  public void flush() {
    if (writeable!=null) {
      writeable.flush();
    }
    else {
    }
  }

  /** Returns the error message of the last occured error. After returning it, the error message will be deleted and cannot be queried again.
      @return the last error message
  */

  public String getLastWriteableError() {
    if (writeable!=null) {
      return writeable.getLastWriteableError();
    }
    else {
      return new String();
    }
  }

  /** Returns wether the MJCWriteable is connected to its target.
      @return wether the MJCWriteable is connected to its target
  */

  public boolean isWriteableConnected() {
    if (writeable!=null) {
      return writeable.isWriteableConnected();
    }
    else {
      return false;
    }
  }

/**	Private function for debugging.
*/

  private void cipherOutput(String s) {
//      System.out.println("<Ciphering> "+s);
  }

}
