/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.ciphering;

import java.math.BigInteger;
import java.util.StringTokenizer;
import com.monica.javacom.util.*;

/**	Implements an RSA (Rivest-Shamir-Adleman) encoder/decoder, which is a recent standard for public-key-ciphering of digital data.
		This object uses an RSAKey as the encoding/decoding key, which is part of an RSAKeyPair, generated by an RSAKeyPairGenerator.
		As the main functionality, the coder gets a String which contains the initial data, and returns the coded representation of this String.
		@see RSAKey
		@see RSAKeyPair
		@see RSAKeyPairGenerator
*/

public class RSA {

	private RSAKey key;
  private MJCConverter converter;

/**	Constructs a new RSA encoder/decoder object.
		@param in_key the RSAKey that the RSA will use to encode/decode
*/

	public RSA(RSAKey in_key) {

    converter=new MJCConverter();
		key=in_key;

	}

/**	Encodes the specified String.
		@param text the input text which has to be encoded
		@return the encoded representation of the input text
		@throws ZeroLengthStringException input string must not be empty
*/

	public String encode(String text) throws ZeroLengthStringException {

		if (text.length()==0) throw new ZeroLengthStringException("RSA: encode: input String must not be empty");

		int blockLength=key.getModulus().bitLength()/8, actual=0, maxPos=text.length();
		StringLista blocks=new StringLista();
		byte[] byteok;
		BigInteger number, codednumber;
		String coded=new String();

                byteok=converter.getBytes(text);
		number=new BigInteger(byteok);
		if (number.compareTo(key.getModulus()) < 0) {
			blocks.add(text.substring(actual));
		}
		else {
			while (maxPos-actual > blockLength)
				blocks.add(text.substring(actual, actual+=blockLength));
			if (maxPos-actual>0)
				blocks.add(text.substring(actual));
		}

		while (!blocks.isEmpty()) {
                        byteok=converter.getBytes(blocks.getFirst());
			number= new BigInteger(byteok);
			codednumber=number.modPow(key.getExponent(), key.getModulus());
			coded=coded.concat(codednumber.toString()+" ");
		}

		return coded;

	}

/**	Decodes the specified String.
		@param text the encoded text which must be the result of the encode() function of an RSA object whith the same key, that this RSA is working with.
		@return the decoded representation of the encoded text, the input text of the encoding algorithm
		@throws ZeroLengthStringException input string must not be empty
*/

	public String decode(String text) throws ZeroLengthStringException {

		if (text.length()==0) throw new ZeroLengthStringException("RSA: decode: input String must not be empty");
		
		int blockLength=key.getModulus().bitLength()/8, actual=0, maxPos=text.length();
		StringLista blocks=new StringLista();
		byte[] codedbyteok;
		BigInteger number, codednumber;
		String coded=new String();
		
		StringTokenizer stringTokenizer=new StringTokenizer(text);
		while (stringTokenizer.hasMoreTokens()) {
		        number=new BigInteger(stringTokenizer.nextToken());
			codednumber=number.modPow(key.getExponent(), key.getModulus());
			codedbyteok=codednumber.toByteArray();
			coded=coded.concat(converter.toString(codedbyteok, 0, codedbyteok.length));
		}
        		
		return coded;	
	
	}
	
/**	Returns the RSAKey of this RSA object.
		@return the RSAKey, that this object is working with
*/	
	
	public RSAKey getKey() {

		return key;

	}

/**	Sets the RSAKey of this object.
		@param in_key the new RSAKey that this RSA object should use
*/

	public void setKey(RSAKey in_key) {

		key=in_key;

	}

}
