/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.ciphering;

import java.math.BigInteger;
import java.util.Random;

/**	Realizes an RSAKeyPairGenerator. This object is used to create a proper RSAKeyPair for using it in an RSA encoder/decoder.
		The RSAKeyPair retruned by the generating function must be splitted, and then passed to the RSA object.
		@see RSAKeyPair
		@see RSAKey
		@see RSA
*/

public class RSAKeyPairGenerator {

	private int keyLength, certainty, publicKey;
	Random random;
	
/** 	Constructs a new RSAKeyPairGenerator. If supplied with a proper Random object, it creates a secure RSAKeyPair.
		@param in_keyLength the keylength of the RSAKey to be generated - default: 256
		@param in_certainty the certainty, that the numbers generated are primes, P=1-2^(-certainty) - default: 100
		@param in_publicKey the suggested publicKey-exponent - default: 65537
		@param in_random the Random object the RSAKeyPairGenerator will use to generate the random numbers
		@throws BadInputValueException the in_keyLength must be between 16 and 65536
		@throws BadInputValueException the in_certainty must be between 1 and 65536
		@throws BadInputValueException the in_publicKey must be odd and greater than 2
*/	
	
	public RSAKeyPairGenerator(int in_keyLength, int in_certainty, int in_publicKey, Random in_random) throws BadInputValueException {

		if ((in_keyLength<16) || (in_keyLength>65536)) throw new BadInputValueException("RSA: keyLength must be between 16 and 65536");
		else keyLength=in_keyLength;
		
		if ((in_certainty<1) || (in_certainty>65536)) throw new BadInputValueException("RSA: certainty must be between 1 and 65536");
		else certainty=in_certainty;
		
		if ((in_publicKey<3) || (in_publicKey%2==0)) throw new BadInputValueException("RSA: publicKey must be odd and greater than 2");
		else publicKey=in_publicKey;
		
		random=in_random;

	}

/**	Generates a new RSAKeyPair with the given keyLength, certainty, publicKey and Random object.
		@return a new RSAKeyPair with the given parameters
*/	
	
	public RSAKeyPair generateRSAKeyPair() {

		BigInteger p = new BigInteger(keyLength, certainty, random);
		BigInteger q = new BigInteger(keyLength, certainty, random);
		BigInteger m = p.multiply(q);

		BigInteger p1 = p.subtract(BigInteger.valueOf(1));
		BigInteger q1 = q.subtract(BigInteger.valueOf(1));
		BigInteger fiem = p1.multiply(q1);

		BigInteger e = BigInteger.valueOf(publicKey);
		while (e.gcd(fiem).intValue() > 1) e = e.add(BigInteger.valueOf(2));

		BigInteger d = solveEquvality(fiem, e);

		return new RSAKeyPair(m, d, e);		

	}	

/** 	Solves an equality neccessay to generate the private key from the public key and the modulus.
		@param a0 parameter 1
		@param a1 parameter 2
		@return the solution
*/
	
	private BigInteger solveEquvality(BigInteger a0, BigInteger a1) {
		
		int i;
		int hossz=a0.bitLength()+1;
		BigInteger a[], q[], n[], e[];
		a=new BigInteger[hossz];
		q=new BigInteger[hossz];
//		n=new BigInteger[hossz];
		e=new BigInteger[hossz];

		a[0]=a0; a[1]=a1;
		q[0]=BigInteger.valueOf(0);
//		n[0]=BigInteger.valueOf(0); n[1]=BigInteger.valueOf(0);
		e[0]=BigInteger.valueOf(0); e[1]=BigInteger.valueOf(0);
		i=1;
		do {
			q[i]=a[i-1].divide(a[i]);
			a[i+1]=a[i-1].remainder(a[i]);
		} while (a[++i].compareTo(BigInteger.valueOf(1)) != 0);

		int lepesszam=i+1;
//		n[2]=BigInteger.valueOf(1); n[3]=BigInteger.valueOf(-1).multiply(q[2]);
		e[2]=BigInteger.valueOf(-1).multiply(q[1]); e[3]=q[1].multiply(q[2]).add(BigInteger.valueOf(1));
		for (i=4; i<lepesszam; i++) {
//			n[i]=n[i-2].subtract(n[i-1].multiply(q[i-1]));
			e[i]=e[i-2].subtract(e[i-1].multiply(q[i-1]));
		}

		if (e[lepesszam-1].compareTo(BigInteger.valueOf(0)) > 0)
			return e[lepesszam-1];
		else
			return e[lepesszam-1].add(a0);
	}

/**	Returns the keylength of the RSAKeyPair to be generated.
		@return the keylength of the RSAKeyPair
*/	
	
	public int getKeyLength() {

		return keyLength;

	}

/**	Returns the certainty of the being prime of the modulus components.
		@return the certainty of the being prime
*/	
	
	public int getCertainty() {

		return certainty;

	}

/**	Returns the suggested public key of the RSAKeyPair to be generated.
		@return the suggested public key of the RSAKeyPair
*/	
	
	public int getPublicKey() {

		return publicKey;

	}

/**	Sets the keylength of the RSAKeyPair to be generated.
		@param in_keyLength the keylength of the RSAKey to be generated - default: 256
		@throws BadInputValueException the in_keyLength must be between 16 and 65536
*/	
	
	public void setKeyLength(int in_keyLength) throws BadInputValueException {

		if ((in_keyLength<16) || (in_keyLength>65536)) throw new BadInputValueException("RSA: keyLength must be between 16 and 65536");
		else keyLength=in_keyLength;

	}

/**	Sets the certainty.
		@param in_certainty the certainty, that the numbers generated are primes, P=1-2^(-certainty) - default: 100
		@throws BadInputValueException the in_certainty must be between 1 and 65536
*/	
	
	public void setCertainty(int in_certainty) throws BadInputValueException {

		if ((in_certainty<1) || (in_certainty>65536)) throw new BadInputValueException("RSA: certainty must be between 1 and 65536");
		else certainty=in_certainty;

	}

/**	Sets the suggested public key exponent.
		@param in_publicKey the suggested publicKey-exponent - default: 65537
		@throws BadInputValueException the in_publicKey must be odd and greater than 2
*/	
	
	public void setPublicKey(int in_publicKey) throws BadInputValueException {

		if ((in_publicKey<3) || (in_publicKey%2==0)) throw new BadInputValueException("RSA: publicKey must be odd and greater than 2");
		else publicKey=in_publicKey;

	}

}
