/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.interfaces;

import com.monica.javacom.interfaces.*;

/** Top level sublayer responsible for the visualization of the connection.
    This layer implements a basic character device with the followin features:
    <ul>
      <li>rectangular screen
      <li>positioning within the screen (gotoXY, left, right, up, down)
      <li>inverse characters
      <li>scrolling
      <li>scrolling region
      <li>insert/delete
    </ul>
*/    
public interface MJCVirtualDisplay extends MJCConfigureable {

  /** Returns the MJC name of the MJCVirtualDisplay.
      @return String: the MJC name of the MJCVirtualDisplay
  */  
  public String getVirtualDisplayName();
  
  /** Returns the MJC version of the MJCVirtualDisplay.
      @return String: the MJC name of the MJCVirtualDisplay
  */            
  public String getVirtualDisplayVersion();
  
  /** Sets the screen size in characters. After resize the screen will
      be cleared. 
      @param int sizeX: the width
      @param int sizeY: the height
  */      
  public void setCSize(int sizeX, int sizeY);
  
  /** Sets the screen width in characters. After resize the screen will
      be cleared.
      @param int sizeX: the width
  */      
  public void setCSizeX(int sizeX);
  
  /** Sets the screen height in characters. After resize the screen will
      be cleared.
      @param int sizeY: the height
  */        
  public void setCSizeY(int sizeY);
  
  /** Returns the screen width in characters.
      @return int: the width
  */  
  public int getCSizeX();
  
  /** Returns the screen height in characters.
      @return int: the height
  */      
  public int getCSizeY();

  /** Clears the contents of the screen. The position of the cursor will
      not be changed.
  */
  public void clear();

  /** Sets the position of the cursor. Indexing begins with 0.
      @param int x: the column
      @param int y: the row
  */      
  public void setCPos(int x, int y);
  
  /** Sets the column of the cursor. Indexing begins with 0.
      @param int x: the column
  */      
  public void setCPosX(int x);
  
  /** Sets the row of the cursors. Indexing begins with 0.
      @param int y: the row
  */      
  public void setCPosY(int y);
  
  /** Returns the current column of the cursor. Indexing begins with 0.
      @return int: the current column
  */      
  public int getCPosX();
  
  /** Returns the current row of the cursor. Indexing begins with 0.
      @return int: the current row
  */      
  public int getCPosY();

  /** Adds a character to the current cursor position. The cursor will
      be adwanced once. The cursor can be positioned one characted right
      off the screen. If there is an add character at that position,
      the cursor will go one line down and to the beginning of the line,
      and the character will be added there. If there are no lines to
      go down, the screen will be scrolled down. The removed line at the top
      will be lost. The cursor cannot leave the scrolling region with this 
      method. Characters will be added with the current inverting status.
      @param char c: the characted to be added
  */  
  public void addC(char c);
  
  /** Adds a String to the current cursor position. This method should do
      <b>s.length()</b> times the functionality of 
      <b>addC(s.charAt(i))</b>.
      @param String s: the String to be added
  */  
  public void addS(String s);
  
  /** Adds a char array to the current cursor position. This method
      should do <b>ca.length</b> times the functionality of 
      <b>addC(ca[i])</b>.
      @param char[] ca: the char array to be added
  */      
  public void addCA(char[] ca);
  
  /** Cursors goes one row down. If the current cursor position is at the bottom
      of the screen, the screen will be scrolled down. The row scrolled
      out will be lost. Horizontal position of the cursor will not be altered.
  */      
  public void newLine();

  /** Cursor goes one column to the left. If the current cursor position is the
      first column then the cursor will go up one row (if possible) and to
      the last column. There won't be any scrolling. The cursor cannot leave
      the scrolling region with this method.
  */      
  public boolean goLeft();
  
  /** Cursor goes <>num</b> times to the left. This method should do 
      <b>num</b> times the functionality of <b>goLeft()</b>.
      @param int num: the number of times to go left
  */      
  public int goLeft(int num);
  
  /** Cursor goes one column to the right. If the current cursor position is
      the last column then the cursor will go down one row (if possible) and
      to the first column. There won't be any scrolling. The cursor cannot
      leave the scrolling region wit this method.
  */      
  public boolean goRight();
  
  /** Cursor goes <b>num</b> times to the right. This method should do
      <b>num</b> times the functionality of <b>goRight()</b>.
      @param int num: the number of times to go right
  */      
  public int goRight(int num);
  
  /** Cursor goes one row up (if possible). There won't be any scrolling. 
      Horizontal position of the cursor will not be altered. Cursor
      cannot leave the scrolling region with this method.
  */      
  public boolean goUp();
  
  /** Cursor goes <b>num</b> times up. This method should do 
      <b>num</b> times the functionality of <b>goUp()</b>.
      @param int num: the number of times to go up
  */      
  public int goUp(int num);
  
  /** Cursor goes one row down (if possible). There won't be any scrolling.
      Horizontal position of the cursor will not be altered. Cursor
      cannot leave the scrolling region with this method.
  */      
  public boolean goDown();
  
  /** Cursor goes <b>num</b> times down. This method should do
      <b>num</b> times the functionality of <b>goDown()</b>.
      @param int num: the number of times to go down.
  */      
  public int goDown(int num);

  /** Sets the inverting status. True means inverted, false normal. 
      @param boolan status: the inverting status
  */      
  public void setInvert(boolean status);

  /** Refreshes the screen contents for the user.
  */
  public void refresh();
  
  /** Erases the contents of the current row from the horizontal cursor 
      position (inclusive) to the last column (inclusive). 
  */      
  public void eraseEOL();
  
  /** Should addC/addCA/addS advance the cursor while adding characters.
      True means yes, false no. If the cursor is positioned outside the
      visible screen area and an addC/addCA/addS method is called,
      the cursor will in either case go one row down and to the first
      column (and a scrolling will be performed if required). By default
      addC/addCA/addS will advance the cursor.
      @param boolean status: should the addC/addCA/addS methods advance 
        the cursor while adding characters.
  */	
  public void moveCursor(boolean status);
  
  /** Sets the scrolling region for the screen. Scrolling region is
      given by the first and last line (both includive, indexing begins
      with 0). If the cursor is inside the scrolling are, it
      cannot leave it with addC/addCA/addS/goLeft/goRight/goUp/goDown/
      newLine/insertLine/insertChar/deleteLine/deleteChar. By
      default the scrolling region is the whole screen.
      @param int firstLine: the first line of the scrolling region
      @param int lastLine: the last line of the scrolling region
  */      
  public void setScrollingRegion(int firstLine, int lastLine);
  
  /** Moves all rows in the scolling region one row down. The first row
      in the scrolling region will get empty. Cursor position will
      not be altered.
  */  
  public void scrollUp();
  
  /** Moves all rows int the scrolling region one row up. The last
      row in the scrolling region will get empty. Cursor position
      will not be altered.
  */      
  public void scrollDown();
  
  /** Inserts <b>lineNum</b> rows at the current cursor position. 
      The current row and the rows below the cursor will be moved 
      <b>lineNum</b> rows down. The last row in the scrolling
      region will get lost during the movement. <b>lineNum</b> empty rows
      will be created beginning with the current row. Cursor will move 
      to the first column of the current row.
      @param int lineNum: the number of rows to be inserted
  */            
  public void insertLine(int lineNum);
  
  /** Deletes <b>lineNum</b> rows at the current cursor position.
      The rows below the deleted rows will be moved upwards. At the bottom
      of the scrolling region new empty lines will be added. Cursor
      will move to the first column of the current row.
      @param int lineNum: the number of rows to be deleted
  */      
  public void deleteLine(int lineNum);
  
  /** Inserts an empty character at the cursor position. Characters after
      the cursor position in the current row will be moved one position
      to the right. The last character in the current row will be lost.
  */      
  public void insertChar();
  
  /** Inserts <b>charNum</b> empty characters at the cursor position.
      This method should do <b>charNum</b> times the functionality of
      <b>insertChar()</b>.
      @param int charNum: number of characters to be inserted
  */      
  public void insertChar(int charNum);
  
  /** Deletes a character at the cursor position. Characters after the
      cursor position in the current row will be moved one position to
      the left. The last character in the current row will be empty.
  */      
  public void deleteChar();
  
  /** Deletes <b>charNum</b> characters at the cursor position.
      This method should do <b>charNum</b> times the functionality
      of <b>deleteChar()</b>.
      @param int charNum: number of characters to be deleted
  */      
  public void deleteChar(int charNum);
}
