/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.manager;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;

/** Middle level layer responsible for the MJC connection between the client and server side.
*/

public class MJCManager implements MJCReader, MJCReadable, MJCWriteable, MJCWriter, MJCConfigureable, Runnable , MJCCloser {

  private static final String NAME="MJCTelnetManager";
  private static final String CLIENTNAME="MJCC";
  private static final String SERVERNAME="MJCS";
  private static final String VERSION="0.1";

  private MJCReadable readable;
  private MJCReader reader;
  private MJCWriteable writeable;
  private MJCConfiguration configuration;

  private boolean threadSuspended;
  private boolean threadSuspendedForReadout;
  private boolean threadRunning;
  private boolean transparentStatus;

  private MJCTransport transport;
  private MJCWindow window;

  private char DELIMITER;

  private MJCStringBuffer readBuffer;
  private int status;

  private MJCConfigureable[] reconf;

  private static final int STAT_STARTED=0;
  private static final int STAT_VALID_SERVER=1;
  private static final int STAT_USER_START=9;
  private static final int STAT_USER=10;
  private static final int STAT_CONF=11;

  /** Default constructor.
  */
  public MJCManager() {
    Thread readerThread=new Thread(this);
    status=STAT_STARTED;
    threadRunning=true;
    threadSuspended=true;
    threadSuspendedForReadout=true;
    readBuffer=new MJCStringBuffer(100000);
    readerThread.start();
  }

  /** Sets the MJCTransport object which is responsible for the data connection between
      the client and the server side.
      @param MJCTransport transport: the MJCTransport that is responsible for the
        data connection between the client and the server side
  */
  public void setTransport(MJCTransport transport) {
    this.transport=transport;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCWriter */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCWriter.
      @return String: the MJC name of the MJCWriter
  */
  public String getWriterName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCWriter.
      @return String: the MJC version of the MJCWriter
  */
  public String getWriterVersion() {
    return VERSION;
  }

  /** Sets the MJCWriteable to which this MJCWriter will write.
      @param MJCWriteable writeable: the MJCWriteable to which this
        MJCWriter will write
  */
  public void setWriteable(MJCWriteable writeable) {
    this.writeable=writeable;
  }

  /** Returns the MJCWriteable to which this MCJWriter will write.
      @return MJCWriteable: the MJCWriteable to which this MJCWriter
        will write
  */
  public MJCWriteable getWriteable() {
    return writeable;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCConfigureable */
  /* ---------------------------------------------------------------------------------- */

  /** Sets the MJCConfiguration for this object.
      @param MJCConfiguration configuration: the MJCConfiguration
        object for the customization
  */
  public void setConfiguration(MJCConfiguration configuration) {
    this.configuration=configuration;
  }

  /** Returns the MJCConfiguration with which this object is customized.
      @return MJCConfiguration: the MJCConfiguration with which this
        object is customized
  */
  public MJCConfiguration getConfiguration() {
    return configuration;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCReadable */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCReadable.
      @return String: the MJC name of the MJCReadable
  */
  public String getReadableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReadable.
      @return String: the MJC version of the MJCReadable
  */
  public String getReadableVersion() {
    return VERSION;
  }

  /** Returns wether the MJCReadable is still connected to its source.
      @return boolean: wether the MJCReadable is still connected to its
        source
  */
  public boolean isReadableConnected() {
    if (readable!=null) {
      return readable.isReadableConnected();
    }
    else {
      return false;
    }
  }

  /** Returns the number of bytes that are available for read out from
      this MJCReadable.
      @return int: the number of bytes available for read out
  */
  public int getAvailable() {
    return readBuffer.getBufferLength();
  }

  /** Returns the contents of the read buffer from this MJCReadable.
      @return String: the contents of the read buffer
  */
  public synchronized String getInput() {
    String temp=readBuffer.emptyBuffer();
    threadSuspendedForReadout=false;
    notify();
    return temp;
  }

  /** Sets the MJCReader to which the readEvents have to be sent.
      @param MJCReader reader: the MJCReader to which the readEvents
        have to be sent
  */
  public void setReader(MJCReader reader) {
    this.reader=reader;
    if (readBuffer.getBufferLength()>0) {
      reader.readEvent(this);
    }
  }

  /** Returns the MJCReader to which readEvents will be sent.
      @return MJCReader: the MJCReader to which readEvents will be sent
  */
  public MJCReader getReader() {
    return reader;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCReader */
  /* ---------------------------------------------------------------------------------- */

  /** Sets the MJCReadable from which data has to be read.
      @param MJCReadable readable: the MJCReadable from which data has
        to be read
  */
  public void setReadable(MJCReadable readable) {
    this.readable=readable;
  }

  /** Returns the MJCReadable from which data will be read.
      @return MJCReadable: the MJCReadable from which data will be read
  */
  public MJCReadable getReadable() {
    return readable;
  }

  /** Entry point for the reader thread.
  */
  public void run() {
    String temp;
    while (threadRunning) {
      synchronized(this) {
        while (threadSuspended) {
          try {
            wait();
          }
          catch (Exception e) {
            debugOutput("run exc I");
          }
        }
        threadSuspended=true;
      }
      temp=readable.getInput();
      if (readBuffer.addToBuffer(temp)) {
        processBuffer();
      }
      else {
        threadSuspended=true;
        threadSuspendedForReadout=true;
        synchronized (this) {
          while (threadSuspendedForReadout) {
            try {
              wait();
            }
            catch (Exception e) {
            }
          }
        }
        readBuffer.addToBuffer(temp);
	      processBuffer();
      }
    }
  }

  /** Returns the MJC name of the MJCReader.
      @return String: the MJC name of the MJCReader
  */
  public String getReaderName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReader.
      @return String: the MJC version of the MJCReader
  */
  public String getReaderVersion() {
    return VERSION;
  }

  /** Indicates that there is data available from the specified
      MJCReadable.
      @param MJCReadable readable: there is data available from the
        readable
  */
  public synchronized void readEvent(MJCReadable readable) {
    threadSuspended=false;
    notify();
  }

  /** Indicates that the source connection of the specified
      MJCReadable has been closed.
      @param MJCReadable readable: the MJCReadable whose source connection
        has been closed
  */
  public synchronized void connectionClosedEvent(MJCReadable readable) {
    if (reader!=null) {
      reader.connectionClosedEvent(this);
    }
    threadRunning=false;
    threadSuspended=false;
    try {
      notify();
    }
    catch (Exception e) {
    }
  }

  /** Indicates that there was an error in the source connection
      of the specified MJCReadable.
      @param String errorDescription: the error message
      @param MJCReadable readable: the MJCReadable where the error
        occured
  */
  public void errorEvent(String errorDescription, MJCReadable readable) {
    if (reader!=null) {
      reader.errorEvent(errorDescription, this);
    }
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCWriteable */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCWriteable.
      @return String: the MJC name of the MJCWriteable
  */
  public String getWriteableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCWriteable.
      @return String: the MJC version of the MJCWriteable
  */
  public String getWriteableVersion() {
    return VERSION;
  }

  /** Writes one single char. Writing is synchronous and can be
      buffered.
      @param char c: the char to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(char c) {
    if (writeable!=null) {
      return writeable.write(c);
    }
    else {
      return false;
    }
  }

  /** Writes a String. Writing is synchronous and can be buffered.
      @param String s: the String to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(String s) {
    if (writeable!=null) {
      return writeable.write(s);
    }
    else {
      return false;
    }
  }

  /** Writes a char array. Writing is synchronous and can be buffered.
      @param char[] c: the char array to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(char[] c) {
    if (writeable!=null) {
      return writeable.write(c);
    }
    else {
      return false;
    }
  }

  /** Flushes the contents of the write buffer to bottom layers.
  */
  public void flush() {
    if (writeable!=null) {
      writeable.flush();
    }
    else {
    }
  }

  /** Returns the error message of the last occured error. After
      returning it, the error message will be deleted and cannot
      be queried again.
      @return String: the last error message
  */
  public String getLastWriteableError() {
    if (writeable!=null) {
      return writeable.getLastWriteableError();
    }
    else {
      return new String();
    }
  }

  /** Returns wether the MJCWriteable is connected to its target.
      @return boolean: wether the MJCWriteable is connected to
        its target
  */
  public boolean isWriteableConnected() {
    if (writeable!=null) {
      return writeable.isWriteableConnected();
    }
    else {
      return false;
    }
  }

  /** This method writes debug information to ste standard outut.
      @patam String s: the debug information to be written to the standard output
  */
  private void debugOutput(String s) {
//    System.err.println("[M]"+s);
  }

  /** This method processes the read buffer, the data stream from the server.
  */
  private synchronized void processBuffer() {
    String buf=readBuffer.getBuffer();

    switch (status) {

      case STAT_STARTED:
        if (buf.length()>0) {
	        DELIMITER=buf.charAt(0);
	        buf=buf.substring(1);
	      }
	      if (buf.length()>0) {
	        int d1pos=buf.indexOf((int)DELIMITER);
	        if (d1pos!=-1) {
	          String sn=buf.substring(0,d1pos);
	          if (sn.equals(SERVERNAME)) {
	            if (d1pos!=(buf.length()-1)) {
	              buf=buf.substring(d1pos+1);
		            int d2pos=buf.indexOf((int)DELIMITER);
		            if (d2pos!=-1) {
		              String sv=buf.substring(0,d2pos);
                  float valueExtern=(Float.valueOf(sv)).floatValue();
		              float valueIntern=(Float.valueOf(VERSION)).floatValue();
                  if (valueIntern<=valueExtern) {
                    readBuffer.emptyBuffer();
                    status=STAT_VALID_SERVER;
                    String s=new String(CLIENTNAME+DELIMITER+VERSION+DELIMITER);
                    debugOutput(s);
                    debugOutput((new Boolean((writeable.write(CLIENTNAME+DELIMITER+VERSION+DELIMITER)))).toString());
                    writeable.flush();
		              }
		              else {
  	                transport.closeConnections();
                    return;
                  }
		            }
	            }
	          }
	          else {
       	      transport.closeConnections();
	            return;
	          }
	        }
	      }
        break;

      case STAT_VALID_SERVER:
        if (buf.equals(new String("SC"+DELIMITER))) {
          readBuffer.emptyBuffer();
          ((MJCCiphering)writeable).setTransparent(false);
          if (reconf!=null) {
            status=STAT_CONF;
            writeable.write("C"+(char)10);
          }
          else {
            status=STAT_USER_START;
            writeable.write("S"+(char)10);
          }
          writeable.flush();
        }
        break;

      case STAT_CONF:
        if (buf.length()>3) {
          if ((buf.substring(0,2)).equals(new String("CS"))) {
            int dpos=buf.indexOf(""+DELIMITER);
            if (dpos!=0) {
              int dnum=Integer.parseInt(buf.substring(2,dpos));
              if ((buf.length()-dpos-1)>=dnum) {
                String dstring=buf.substring(dpos+1, dpos+1+dnum);
                String drest=buf.substring(dpos+1+dnum);
                readBuffer.clearBuffer();
                readBuffer.addToBuffer(drest);
                MJCConfiguration strConf=new MJCStringConfiguration(dstring);
                for (int i=0; i<reconf.length; i++) reconf[i].setConfiguration(strConf);
                status=STAT_USER_START;
                writeable.write("S"+(char)10);
                writeable.flush();
              }
            }
          }
        }
        break;

      case STAT_USER_START:
        if (buf.length()>2) {
          if ((buf.substring(0,3)).equals(new String("SU"+DELIMITER))) {
            buf=buf.substring(3);
            readBuffer.clearBuffer();
            readBuffer.addToBuffer(buf);
            status=STAT_USER;
            if (reader!=null) {
              reader.readEvent(this);
            }
          }
        }
        break;

      case STAT_USER:
        if (reader!=null) {
          reader.readEvent(this);
        }
        break;
    }
  }

  /** Returns the MJC name of the MJCCloser.
      @return String: the MJC name of the MJCCloser
  */
  public String getCloserName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCCloser.
      @return String: the MJC version of the MJCCloser
  */
  public String getCloserVersion() {
    return VERSION;
  }

  /** This method closes the client side application.
  */
  public void closeApplication() {
    transport.closeConnections();
  }

  /** Sets the MJCWindow that represents the user interface of the client
      side.
      @param MJCWindow window: the MJCWindow that represents the
        user interface of the client side
  */
  public void setWindow(MJCWindow window) {
    this.window=window;
  }

  public void setReconfigureable(MJCConfigureable[] r) {
    reconf=r;
  }
}
