/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.protocol;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import com.monica.javacom.ciphering.*;
import com.monica.javacom.transport.*;
import java.util.*;
import java.io.*;

/** Class for implementing a basic loopback at the server side. This class acts as a top level
    layer int the MJC hierarchy on the server side. Accepted connections will be forwarded to
    the destination host.
*/
public class MJCLoopBackProtocol implements MJCServerManager, MJCReader, Runnable {

  private static final String NAME="MJCLoopBackProtocol";
  private static final String VERSION="0.1";
  private static final String SERVER_NAME="MJCS";
  private static final String SERVER_VERSION="0.1";
  private static final String CLIENT_NAME="MJCC";
  private static final char DELIMITER=(char)35;

  private static final int tempReadBufferLength=100;

  private static final int DEBUG_NO=0;
  private static final int DEBUG_TOTALS=1;
  private static final int DEBUG_TRANSFER=2;
  private static final int DEBUG_DETAILED=3;

  private static final int STAT_NOTVALID=0;
  private static final int STAT_VALID_COMMAND=1;
  private static final int STAT_VALID_USER=2;

  private int workPort;
  private int adminPort;
  private String forwardToHost;
  private int forwardToPort;
  private int port;
  private MJCLoopBackProtocol loopWork, loopAdmin, owner;
  private MJCConfiguration configuration;
  private Thread threadWork, threadAdmin;
  private MJCTransport transportWork, transportAdmin;

  private Vector inTransports, outTransports, inCipherings;
  private Vector clientStatus, tempReadBuffers, tempReadBufferPos;

  private String configInput;

  private MJCConverter converter;

  private MJCTransport transport;

  /** Sets the customization file that will be sent through the network for clients to
      allow them basic configuration via the server.
      @param String is: the location of the customization file
  */
  public void setConfigurationInput(String is) {
    configInput=is;
  }

  /** Default constructor. Workport (where the connections will be accepted), the name of the
      host to forward to, and the destination port as well as the admin port have to be
      specified.
      @param int workPort: the port number where the clients connect
      @param String forwardToHost: the host to forward to
      @param int forwardToPort: the port to forward to
      @param int adminPort: the administration port
  */
  public MJCLoopBackProtocol(int workPort, String forwardToHost, int forwardToPort, int adminPort) {
    this.workPort=workPort;
    this.adminPort=adminPort;
    this.forwardToHost=forwardToHost;
    this.forwardToPort=forwardToPort;
    inTransports=new Vector();
    inCipherings=new Vector();
    outTransports=new Vector();
    clientStatus=new Vector();
    tempReadBuffers=new Vector();
    tempReadBufferPos=new Vector();
    converter=new MJCConverter();
  }

  /** Private constructor for creating a new thread for each client.
      @param MJCTransport transport: the MJCTransport for the client
      @param int port: the port number of the server
      @param MJCLoopBackProtocol owner: the owner of the threads
  */
  private MJCLoopBackProtocol(MJCTransport transport, int port, MJCLoopBackProtocol owner) {
    this.transport=transport;
    this.port=port;
    this.owner=owner;
    transport.setConfiguration(configuration);
    transport.setServerManager(owner);
    inTransports=new Vector();
    inCipherings=new Vector();
    outTransports=new Vector();
    converter=new MJCConverter();
  }

  /** This method starts the application.
  */
  public void start() {
    transportWork=new MJCTCPTransport();
    transportAdmin=new MJCTCPTransport();
    loopWork=new MJCLoopBackProtocol(transportWork, workPort, this);
    loopWork.setConfiguration(configuration);
    loopAdmin=new MJCLoopBackProtocol(transportAdmin, adminPort, this);
    loopAdmin.setConfiguration(configuration);
    threadWork=new Thread(loopWork);
    threadAdmin=new Thread(loopAdmin);
    threadWork.start();
    threadAdmin.start();
  }

  /** Entry point for the thread for each client.
  */
  public void run() {
    try {
      transport.initAsServer(port);
    }
    catch (Exception e) {
      debugOutput("lbp.run exc"+e.toString());
      System.exit(0);
    }
  }

  /** Sets the MJCConfiguration for this object.
      @param MJCConfiguration configuration: the MJCConfiguration
        object for the customization
  */
  public void setConfiguration(MJCConfiguration configuration) {
    this.configuration=configuration;
  }

  /** Returns the MJCConfiguration with which this object is customized.
      @return MJCConfiguration: the MJCConfiguration with which this
        object is customized
  */
  public MJCConfiguration getConfiguration() {
    return configuration;
  }

  /** Returns the MJC name of the MJCServerManager.
      @return String: the MJC name of the MJCServerManager
  */
  public String getServerManagerName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCServerManager.
      @return String: the MJC version of the MJCServerManager
  */
  public String getServerManagerVersion() {
    return VERSION;
  }

  /** This method will be invoked from the MJCTransport when a new
      client arrives.
      @param MJCTransport server: the server that has accepted the client
      @param MJCTrabsport client: the accepted client
  */
  public synchronized void newClientHasArrivedEvent(MJCTransport server, MJCTransport client) {
    MJCTransport t=new MJCTCPTransport();
    MJCCiphering c=new MJCBasicCiphering(true);
    if (server==transportWork) {
      debugOutput("New client has arrived on the work port.");
      t.setConfiguration(configuration);
      c.setConfiguration(configuration);
      client.setConfiguration(configuration);

      c.setReadable(client);
      c.setReader(this);
      c.setWriteable(client);

      client.setReader(c);

      t.setReader(this);

      inTransports.addElement(client);
      inCipherings.addElement(c);
      outTransports.addElement(t);

      clientStatus.addElement(new Integer(STAT_NOTVALID));

      tempReadBuffers.addElement(new char[tempReadBufferLength]);
      tempReadBufferPos.addElement(new Integer(0));

      sendInit(c, STAT_NOTVALID);

    }
    else if (server==transportAdmin) {
      System.exit(0);
    }
  }

  /** Starts forwarding for the client <b>pos</p>.
      @param int pos: the client for that the forwarding has to be started
  */
  private void startForwarding(int pos) {
    MJCTransport t=(MJCTransport)(outTransports.elementAt(pos));
    try {
      t.initAsClient(forwardToHost, forwardToPort);
    }
    catch (Exception e) {
      debugOutput("could not start forwarding.");
      t.closeConnections();
    }
  }

  /** Returns the MJC name of the MJCReader.
      @return String: the MJC name of the MJCReader
  */
  public String getReaderName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReader.
      @return String: the MJC version of the MJCReader
  */
  public String getReaderVersion() {
    return VERSION;
  }

  /** Indicates that there is data available from the specified
      MJCReadable.
      @param MJCReadable readable: there is data available from the
        readable
  */
  public void readEvent(MJCReadable readable) {
    int inPos=inCipherings.indexOf(readable);
    int outPos=outTransports.indexOf(readable);

    String temp;
    if (inPos!=-1) {
      temp=((MJCCiphering)inCipherings.elementAt(inPos)).getInput();
      int status=((Integer)(clientStatus.elementAt(inPos))).intValue();
      switch (status) {
        case STAT_NOTVALID :
          processNotValid(temp, inPos);
          break;
        case STAT_VALID_COMMAND :
          processValidCommand(temp, inPos);
          break;
        case STAT_VALID_USER :
          if (!((MJCTransport)outTransports.elementAt(inPos)).write(temp)) {
            // hibakezeles
          }
          else {
            ((MJCTransport)outTransports.elementAt(inPos)).flush();
          }
          break;
      }
    }
    else if (outPos!=-1) {
      temp=((MJCTransport)outTransports.elementAt(outPos)).getInput();
      if (!((MJCCiphering)inCipherings.elementAt(outPos)).write(temp)) {
        // hibakezeles
      }
      else {
        ((MJCCiphering)inCipherings.elementAt(outPos)).flush();
      }
    }
  }

  /** Indicates that the source connection of the specified
      MJCReadable has been closed.
      @param MJCReadable readable: the MJCReadable whose source connection
        has been closed
  */
  public void connectionClosedEvent(MJCReadable readable) {
    int inPos=inCipherings.indexOf(readable);
    int outPos=outTransports.indexOf(readable);

    MJCTransport t=null;
    debugOutput("connectionClosedEvent inPos="+inPos+" outPos="+outPos);

    if (inPos!=-1) {
      t=(MJCTransport)outTransports.elementAt(inPos);
      synchronized (this) {
        removeClient(inPos);
      }
      t.closeConnections();
    }
    else if (outPos!=-1) {
      t=(MJCTransport)inTransports.elementAt(outPos);
      synchronized (this) {
        removeClient(outPos);
      }
      t.closeConnections();
    }
  }

  /** Removes the client with <b>pos</b> from the list of clients handles.
      @param int pos: the number of the client to be removed
  */
  private void removeClient(int pos) {
    inTransports.removeElementAt(pos);
    outTransports.removeElementAt(pos);
    inCipherings.removeElementAt(pos);
    clientStatus.removeElementAt(pos);
    tempReadBuffers.removeElementAt(pos);
    tempReadBufferPos.removeElementAt(pos);
  }

  /** Indicates that there was an error in the source connection
      of the specified MJCReadable.
      @param String errorDescription: the error message
      @param MJCReadable readable: the MJCReadable where the error
        occured
  */
  public void errorEvent(String errorDescription, MJCReadable readable) {
    int inPos=inCipherings.indexOf(readable);
    int outPos=outTransports.indexOf(readable);
    debugOutput("errorEvent inPos="+inPos+", outPos="+outPos);
    // hibakezeles
  }

  /** Sets the MJCReadable from which data has to be read.
      @param MJCReadable readable: the MJCReadable from which data has
        to be read
  */
  public void setReadable(MJCReadable readable) {
  }

  /** Returns the MJCReadable from which data will be read.
      @return MJCReadable: the MJCReadable from which data will be read
  */
  public MJCReadable getReadable() {
    return null;
  }

  /** Sends debug information to the standard out.
      @param String s: the debug information
  */
  private void debugOutput(String s) {
    System.out.println("[LBP]"+s);
  }

  /** Processes read buffers for not validated clients.
      @param String s: the read buffer
      @param int inPos: the number of the client
  */
  private void processNotValid(String s, int inPos) {
    char[] ca;
    for (int i=0; i<s.length(); i++) {
      ca=addToTempReadBuffer(s.charAt(i),inPos);
      if (ca!=null) {
        try {
          if (validateClient(ca, inPos)) {
            clientStatus.setElementAt(new Integer(STAT_VALID_COMMAND), inPos);
            debugOutput("valid client");
            MJCCiphering c=(MJCCiphering)(inCipherings.elementAt(inPos));
            sendInit(c,STAT_VALID_COMMAND);
            break;
          }
        }
        catch (Exception e) {
          connectionClosedEvent((MJCTransport)(outTransports.elementAt(inPos)));
          return;
        }
      }
    }
  }

  /** Processes MJC commands for validated clients.
      @param String s: the read buffer
      @param int inPos: the number of the client
  */
  private void processValidCommand(String s, int inPos) {
    switch (s.charAt(0)) {
      case 'S' :
        clientStatus.setElementAt(new Integer(STAT_VALID_USER),inPos);
        sendInit(((MJCCiphering)(inCipherings.elementAt(inPos))), STAT_VALID_USER);
        startForwarding(inPos);
        break;
      case 'C' :
        try {
          FileInputStream configInputStream=new FileInputStream(configInput);
          int i=configInputStream.available();
          ((MJCCiphering) (inCipherings.elementAt(inPos))).write("CS"+i+DELIMITER);
          byte[] b=new byte[i];
          int j=configInputStream.read(b);
          if (j!=i) throw new Exception();
          ((MJCCiphering) (inCipherings.elementAt(inPos))).write(converter.toString(b));
          ((MJCCiphering)(inCipherings.elementAt(inPos))).flush();
        }
        catch (Exception e) {
          debugOutput(e.toString());
        }
        break;
    }
  }

  /** Adds a char to the temporary read buffer for a specified client.
      @param char c: the char to be added
      @param int pos: the number of the client for which this buffer has to be altered
      @return char[]: the temporary read buffer for the client
  */
  private char[] addToTempReadBuffer(char c, int i) {
    char[] ca=(char[])(tempReadBuffers.elementAt(i));
    int pos=((Integer)(tempReadBufferPos.elementAt(i))).intValue();
    if (pos<tempReadBufferLength) {
      ca[pos]=c;
      pos++;
      tempReadBufferPos.setElementAt(new Integer(pos),i);
      tempReadBuffers.setElementAt(ca,i);
      return ca;
    }
    return null;
  }

  /** Validates the client upon the temporary read buffer. If the client is surely valid, this
      method returns true. If the validation of the client cannot be checked surely based
      on the read buffer, false is returned. If the client is surely invalid, an Exception is thrown.
      @param char[] c: the temporary read buffer for the client
      @param int inPos: the number of the client
      @return boolean: the validation status of the client
      @exception Exception: thrown if the client is surely invalid
  */
  private boolean validateClient(char[] c, int inPos) throws Exception {
    int mp;
    int pos=((Integer)(tempReadBufferPos.elementAt(inPos))).intValue()-1;

    if (pos<(CLIENT_NAME.length()+2+1)) {
      return false;
    }

    if (!(new String(c,0,CLIENT_NAME.length())).equals(CLIENT_NAME)) {
      throw new Exception();
    }

    mp=-1;
    for (int j=(CLIENT_NAME.length()+1); j<=pos; j++) {
      if (c[j]==DELIMITER) {
        mp=j;
        break;
      }
    }

    if (mp==-1) {
      return false;
    }

    float valueIntern, valueExtern;

    try {
      valueExtern=(Float.valueOf(new String(c,CLIENT_NAME.length()+1,mp-1-CLIENT_NAME.length()))).floatValue();
      valueIntern=(Float.valueOf(SERVER_VERSION)).floatValue();

      if (valueIntern<=valueExtern) {
        return true;
      }
    }
    catch (Exception e) {
      debugOutput("vC exc "+e.toString());
    }

    throw new Exception();
  }

  /** Sends a MJC initialization Strning to the client showing it its status.
      @param MJCWriteable w: the MCJWriteable handling the connection to the client
      @param int status: what status has to be shown
  */
  private void sendInit(MJCWriteable w, int status) {
    switch (status) {
      case STAT_NOTVALID :
        w.write(DELIMITER);
        w.write(SERVER_NAME);
        w.write(DELIMITER);
        w.write(SERVER_VERSION);
        w.write(DELIMITER);
        w.flush();
        break;
      case STAT_VALID_COMMAND :
        w.write("SC"+DELIMITER);
        w.flush();
        break;
      case STAT_VALID_USER :
        w.write("SU"+DELIMITER);
        w.flush();
        break;
    }
  }

}
