/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.protocol;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;

/** Implements TELNET protocol functionality in the MJC layer hierarchy. See the TELNET description
    in RFC 854. This class does currently implement the following TELNET commands/functionality:
    <ul>
      <li>WILL, WONT, DO, DONT
      <li>SB, SE
      <li>Telnet Option - Terminal Type, RFC 1091
      <li>Telnet Option - Terminal Speed, RFC 1079
      <li>Telnet Option - Negotiate About Window Size, RFC 1073
      <li>Telnet Option - Echo, RFC 857
    </ul>
    Writing in this class is not buffered, but reading is buffered.<br><br>
    configureable values
    <ul>
      <li><b>telnetReadBufferSize</b>: maximum size of the read buffer
      <li><b>telnetWriteBufferSize</b>: maximum size of the read buffer
      <li><b>telnetReadSpeed</b>: read speed for Terminal Speed, see RFC 1079
      <li><b>telnetWriteSpeed</b>: write speed for Terminal Speed, see RFC 1079
      <li><b>terminalName</b>: terminal name, see RFC 1091
      <li><b>displayWidth</b>: display width in characters, see RFC 1073
      <li><b>didplayHeight</b>: display height in characters, see RFC 1073
    </ul>
*/
public class MJCTelnetProtocol implements MJCProtocol {

  // ***************************************************************************************
  // VARIABLES
  // ***************************************************************************************
  // Communication
  MJCReader reader;
  MJCWriteable writeable;
  MJCReadable readable;

  // Buffers
  String readBuffer;
  char[] writeBuffer;
  int writeBufferPos;
  int writeBufferSize;
  int readBufferPos;
  String writeBufferString;
  MJCStringBuffer readableBuffer;

  // Telnet status
  boolean ts_control;
  boolean ts_do;
  boolean ts_will;
  boolean ts_sb;
  boolean ts_unsupportedSb;
  boolean ts_dont;

  // Program configuration
  MJCConfiguration configuration;

  boolean processGo;
  char act;

  // ***************************************************************************************
  // TELNET PROTOCOL COMMAND CODES
  // ***************************************************************************************
    private static final int MJC_TELOPT_NULL=0;
    private static final int MJC_TELOPT_LF=10; /* Line feed*/
    private static final int MJC_TELOPT_CR=13; /* Carriage return */
    private static final int MJC_TELOPT_BELL=7;
    private static final int MJC_TELOPT_BACKSPACE=8;
    private static final int MJC_TELOPT_HT=9; /* Horizontal tab */
    private static final int MJC_TELOPT_VT=11; /* Vertical tab */
    private static final int MJC_TELOPT_FF=12; /* Form feed*/
    private static final int MJC_TELOPT_XON=17;
    private static final int MJC_TELOPT_XOFF=19;
    private static final int MJC_TELOPT_SE=240; /* End of subnegotiation parameters */
    private static final int MJC_TELOPT_NOP=241; /* No operation*/
    private static final int MJC_TELOPT_DM=242; /* Data mark */
    private static final int MJC_TELOPT_BRK=243; /* Break */
    private static final int MJC_TELOPT_IP=244; /* Interrupt process */
    private static final int MJC_TELOPT_AO=245; /* Abort output */
    private static final int MJC_TELOPT_AYT=246; /* Are you there*/
    private static final int MJC_TELOPT_EC=247; /* Erase character */
    private static final int MJC_TELOPT_EL=248; /* Erase line */
    private static final int MJC_TELOPT_GA=249; /* Go ahead*/
    private static final int MJC_TELOPT_SB=250; /* Start of subnegotiation */
    private static final int MJC_TELOPT_WILL=251;
    private static final int MJC_TELOPT_WONT=252;
    private static final int MJC_TELOPT_DO=253;
    private static final int MJC_TELOPT_DONT=254;
    private static final int MJC_TELOPT_IAC=255;

  // ***************************************************************************************
  // TELNET OPTION COMMAND CODES
  // ***************************************************************************************
    private static final int MJC_TELOPT_TRANSMIT_BINARY=0; /* RFC 856 */
    private static final int MJC_TELOPT_ECHO=1; /* RFC 857 */
    private static final int MJC_TELOPT_SUPPRESS_GO_AHEAD=3; /* RFC 858 */
    private static final int MJC_TELOPT_STATUS=5; /* RFC 859 */
    private static final int MJC_TELOPT_TIMING_MARK=6; /* RFC 860 */
    private static final int MJC_TELOPT_RCTE=7; /* Remote Controlled Transmission and Echoing, RFC 726 */
    private static final int MJC_TELOPT_NAOCRD=10; /* Negotiate About Output Carriage Return Disposition, RFC 652 */
    private static final int MJC_TELOPT_NAOHTS=11; /* Negotiate About Output Horizontal Tabstops, RFC 653 */
    private static final int MJC_TELOPT_NAOHTD=12; /* Negotiate About Output Horizontal Tab Disposition, RFC 654 */
    private static final int MJC_TELOPT_NAOFFD=13; /* Negotiate About Output Formfeed Disposition, RFC 655 */
    private static final int MJC_TELOPT_NAOVTS=14; /* Negotiate About Vertical Tabstops, RFC 656 */
    private static final int MJC_TELOPT_NAOVTD=15; /* Negotiate About Output Vertical Tab Disposition, RFC 657 */
    private static final int MJC_TELOPT_NAOLFD=16; /* Negotiate About Output Linefeed Disposition, RFC 658 */
    private static final int MJC_TELOPT_EXTEND_ASCII=17; /* RFC 698 */
    private static final int MJC_TELOPT_LOGOUT=18; /* RFC 727 */
    private static final int MJC_TELOPT_BM=19; /* Byte Macro, RFC 735*/
    private static final int MJC_TELOPT_DET=20; /* Data Terminal Entry, DODIIS, RFC 1043 */
    private static final int MJC_TELOPT_SUPDUP=21; /* RFC 736 */
    private static final int MJC_TELOPT_SUPDUT_OUPUT=22; /* RFC 749 */
    private static final int MJC_TELOPT_SEND_LOCATION=23; /* RFC 779 */
    private static final int MJC_TELOPT_TERMINAL_TYPE=24; /* RFC 1091 */
    private static final int MJC_TELOPT_END_OF_RECORD=25; /* RFC 885 */
    private static final int MJC_TELOPT_TUID=26; /* TACACS User Identification, RFC 927 */
    private static final int MJC_TELOPT_OUTMRK=27; /* Output Marking, RFC 933 */
    private static final int MJC_TELOPT_TTYLOC=28; /* Terminal Location Number, RFC 946 */
    private static final int MJC_TELOPT_3270_REGIME=29; /* RFC 1041 */
    private static final int MJC_TELOPT_X3PAD=30; /* RFC 1053 */
    private static final int MJC_TELOPT_NAWS=31; /* Negotiate About Window Size, RFC 1073 */
    private static final int MJC_TELOPT_TERMINAL_SPEED=32; /* RFC 1079 */
    private static final int MJC_TELOPT_TOGGLE_FLOW_CONTROL=33; /* RFC 1372 */
    private static final int MJC_TELOPT_LINEMODE=34; /* RFC 1184 */
    private static final int MJC_TELOPT_XDISP_LOC=35; /* X Display Location RFC 1096 */
    private static final int MJC_TELOPT_NEW_ENVIRON=39; /* RFC 1572 */
    private static final int MJC_TELOPT_TN3270E=40; /* RFC 1647*/
    private static final int MJC_TELOPT_EXOPL=255; /* Extend Options, List Option RFC 861 */

  // ***************************************************************************************
  // CONFIGURATION FILE CONSTANTS
  // ***************************************************************************************

    private static final String CONF_READ_BUFFER_SIZE="telnetReadBufferSize";
    private static final String CONF_WRITE_BUFFER_SIZE="telnetWriteBufferSize";

    private static final String CONF_TERMINAL_SPEED_READ="telnetReadSpeed";
    private static final String CONF_TERMINAL_SPEED_WRITE="telnetWriteSpeed";

    private static final String CONF_TERMINAL_NAME="terminalName";

    private static final String CONF_DISPLAY_SIZE_WIDTH="displayWidth";
    private static final String CONF_DISPLAY_SIZE_HEIGHT="didplayHeight";

    private static final int P_READ_BUFFER_DEF_SIZE=50000;
    private static final int P_WRITE_BUFFER_DEF_SIZE=50000;

    private static final String P_TERMINAL_SPEED_DEF_READ=new String("38400");
    private static final String P_TERMINAL_SPEED_DEF_WRITE=new String("38400");

    private static final String P_TERMINAL_NAME_DEF="vt220";

    private static final int P_DISPLAY_SIZE_DEF_WIDTH=80;
    private static final int P_DISPLAY_SIZE_DEF_HEIGHT=25;

  // ***************************************************************************************
  // NAMING
  // ***************************************************************************************

    private static final String NAME="MJCTelnetProtocol";
    private static final String VERSION="0.1";

  // ***************************************************************************************
  // MJCProtocol
  // ***************************************************************************************

  /** Returns the MJC name of the MJCProtocol.
      @return String: the MJC name of the MJCProtocol
  */
  public String getProtocolName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCProtocol.
      @return String: the MJC version of the MJCProtocol
  */
  public String getProtocolVersion() {
    return VERSION;
  }

  // ***************************************************************************************
  // MJCWriter
  // ***************************************************************************************

  /** Returns the MJC name of the MJCWriter.
      @return String: the MJC name of the MJCWriter
  */
  public String getWriterName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCWriter.
      @return String: the MJC version of the MJCWriter
  */
  public String getWriterVersion() {
    return VERSION;
  }

  /** Sets the MJCWriteable to which this MJCWriter will write.
      @param MJCWriteable writeable: the MJCWriteable to which this
        MJCWriter will write
  */
  public void setWriteable(MJCWriteable writeable) {
    this.writeable=writeable;
  }

  /** Returns the MJCWriteable to which this MCJWriter will write.
      @return MJCWriteable: the MJCWriteable to which this MJCWriter
        will write
  */
  public MJCWriteable getWriteable() {
    return writeable;
  }

  // ***************************************************************************************
  // MJCWriteable
  // ***************************************************************************************

  /** Returns the MJC name of the MJCWriteable.
      @return String: the MJC name of the MJCWriteable
  */
  public String getWriteableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCWriteable.
      @return String: the MJC version of the MJCWriteable
  */
  public String getWriteableVersion() {
    return VERSION;
  }

  /** Writes one single char. Writing is synchronous and can be
      buffered.
      @param char c: the char to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(char c) {
    if (writeable!=null) {
      return writeable.write(c);
    }
    return false;
  }

  /** Writes a String. Writing is synchronous and can be buffered.
      @param String s: the String to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(String s) {
    addToWriteBuffer(s);
    sendWriteBuffer();
    return true;
  }

  /** Writes a char array. Writing is synchronous and can be buffered.
      @param char[] c: the char array to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(char[] c) {
    if (writeable!=null) {
      return writeable.write(c);
    }
    return false;
  }

  /** Flushes the contents of the write buffer to bottom layers.
  */
  public void flush() {
    if (writeable!=null) {
      writeable.flush();
    }
  }

  /** Returns the error message of the last occured error. After
      returning it, the error message will be deleted and cannot
      be queried again.
      @return String: the last error message
  */
  public String getLastWriteableError() {
    if (writeable!=null) {
      return writeable.getLastWriteableError();
    }
    return null;
  }

  /** Returns wether the MJCWriteable is connected to its target.
      @return boolean: wether the MJCWriteable is connected to
        its target
  */
  public boolean isWriteableConnected() {
    if (writeable!=null) {
      return writeable.isWriteableConnected();
    }
    return false;
  }

  // ***************************************************************************************
  // MJCReader
  // ***************************************************************************************

  /** Returns the MJC name of the MJCReader.
      @return String: the MJC name of the MJCReader
  */
  public String getReaderName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReader.
      @return String: the MJC version of the MJCReader
  */
  public String getReaderVersion() {
    return VERSION;
  }

  /** Indicates that there is data available from the specified
      MJCReadable.
      @param MJCReadable readable: there is data available from the
        readable
  */
  public void readEvent(MJCReadable readable) {
    if (readable!=null) {
      while (readable.getAvailable()>0) {
        process(readable.getInput());
      }
    }
  }

  /** Indicates that the source connection of the specified
      MJCReadable has been closed.
      @param MJCReadable readable: the MJCReadable whose source connection
        has been closed
  */
  public void connectionClosedEvent(MJCReadable readable) {
    if (reader!=null) {
      reader.connectionClosedEvent(this);
    }
  }

  /** Indicates that there was an error in the source connection
      of the specified MJCReadable.
      @param String errorDescription: the error message
      @param MJCReadable readable: the MJCReadable where the error
        occured
  */
  public void errorEvent(String errorDescription, MJCReadable readable) {
    if (reader!=null) {
      reader.errorEvent(errorDescription,this);
    }
  }

  /** Sets the MJCReadable from which data has to be read.
      @param MJCReadable readable: the MJCReadable from which data has
        to be read
  */
  public void setReadable(MJCReadable readable) {
    this.readable=readable;
  }

  /** Returns the MJCReadable from which data will be read.
      @return MJCReadable: the MJCReadable from which data will be read
  */
  public MJCReadable getReadable() {
    return readable;
  }

  // ***************************************************************************************
  // MJCConfigureable
  // ***************************************************************************************

  /** Sets the MJCConfiguration for this object.
      @param MJCConfiguration configuration: the MJCConfiguration
        object for the customization
  */
  public void setConfiguration(MJCConfiguration configuration) {
    this.configuration=configuration;
    initValuesFromConfig();
  }

  /** Returns the MJCConfiguration with which this object is customized.
      @return MJCConfiguration: the MJCConfiguration with which this
        object is customized
  */
  public MJCConfiguration getConfiguration() {
    return configuration;
  }

  // ***************************************************************************************
  // MJCReadable
  // ***************************************************************************************

  /** Returns the MJC name of the MJCReadable.
      @return String: the MJC name of the MJCReadable
  */
  public String getReadableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReadable.
      @return String: the MJC version of the MJCReadable
  */
  public String getReadableVersion() {
    return VERSION;
  }

  /** Returns the number of bytes that are available for read out from
      this MJCReadable.
      @return int: the number of bytes available for read out
  */
  public int getAvailable() {
    return readableBuffer.getBufferLength();
  }

  /** Returns the contents of the read buffer from this MJCReadable.
      @return String: the contents of the read buffer
  */
  public String getInput() {
    return readableBuffer.emptyBuffer();
  }

  /** Returns wether the MJCReadable is still connected to its source.
      @return boolean: wether the MJCReadable is still connected to its
        source
  */
  public boolean isReadableConnected() {
    if (readable!=null) {
      return readable.isReadableConnected();
    }
    return false;
  }

  /** Sets the MJCReader to which the readEvents have to be sent.
      @param MJCReader reader: the MJCReader to which the readEvents
        have to be sent
  */
  public void setReader(MJCReader reader) {
    this.reader=reader;
  }

  /** Returns the MJCReader to which readEvents will be sent.
      @return MJCReader: the MJCReader to which readEvents will be sent
  */
  public MJCReader getReader() {
    return reader;
  }

  // ***************************************************************************************
  // CONSTRUCTORS
  // ***************************************************************************************
  /** Contstructor
  */
  public MJCTelnetProtocol() {
    // initializing telnet option variables
    ts_control=false;
    ts_do=false;
    ts_sb=false;
    ts_will=false;
    ts_dont=false;
    ts_unsupportedSb=false;
  }

  // ***************************************************************************************
  // BUFFERING
  // ***************************************************************************************
  /** Buffer management - write to buffer
      @param char c: the char to write to the write buffer
  */
  private void addToWriteBuffer(char c) {
    writeBuffer[writeBufferPos]=c;
    writeBufferPos++;
    if (writeBufferPos==writeBufferSize) {
      sendWriteBuffer();
    }
  }

  /** Buffer management - write to buffer
      @param int i: the in to be written to the write buffer
  */
  private void addToWriteBuffer(int i) {
    writeBuffer[writeBufferPos]=(char)i;
    writeBufferPos++;
    if (writeBufferPos==writeBufferSize) {
      sendWriteBuffer();
    }
  }

  /** Buffer management - write to buffer
      @param String s: the string to be written to the write buffer
  */
  private void addToWriteBuffer(String s) {
    for (int i=0; i<s.length(); i++) {
      writeBuffer[writeBufferPos]=s.charAt(i);
      writeBufferPos++;
      if (writeBufferPos==writeBufferSize) {
        sendWriteBuffer();
      }
    }
  }

  /** Buffer management - send write buffer, ends the write buffer
  */
  private void sendWriteBuffer() {
    try {
      writeBufferString=new String(writeBuffer,0,writeBufferPos);
      writeBufferPos=0;
      send(writeBufferString);
      writeBufferString=null;
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** How many chars are still available in the read buffer
      @return int: how many chars are still available in the read buffer
  */
  private int readBufferAvailable() {
    return (readBuffer.length()-readBufferPos);
  }

  // ***************************************************************************************
  // INPUT PROCESSING
  // ***************************************************************************************

  /** Input buffer main processing
      @param String s: the input buffer
  */
  private void process(String s) {

    processGo=true;
    readBuffer=readBuffer.concat(s);
    if (readBuffer.length()==0) {
      processGo=false;
    }

    while (processGo) {
      act=readBuffer.charAt(readBufferPos);
      readBufferPos++;
      switch ((int)act) {
        case MJC_TELOPT_IAC : // beginning command
          ts_control=true;
          break;
        case MJC_TELOPT_DO : // do
          if (ts_control) {
            ts_do=true;
          }
          break;
        case MJC_TELOPT_WILL : // will
          if (ts_control) {
            ts_will=true;
          }
          break;
        case MJC_TELOPT_SB : // start subnegotiation
          if (ts_control) {
            ts_sb=true;
          }
          break;
        case MJC_TELOPT_SE : // end of subnegotiation
          if (ts_unsupportedSb) {
            ts_unsupportedSb=false;
            ts_control=false;
          }
          break;
        case MJC_TELOPT_DONT : // dont
          if (ts_control) {
            ts_dont=true;
          }
          break;
        default :
          if (ts_control) { // control commands
            if (ts_do) { // request for do
              processDo((int)act);
              ts_do=false;
              ts_control=false;
            }
            else if (ts_sb) { // start of subnegotiation
              processSb((int)act);
              ts_sb=false;
              ts_control=false;
            }
            else if (ts_will) { // request for will
              processWill((int)act);
              ts_will=false;
              ts_control=false;
            }
            else if (ts_dont) { // request for dont
//              processWill((int)act);
              ts_dont=false;
              ts_control=false;
            }
            else {
            }
          }
          else if (ts_unsupportedSb) {
          }
          else {
	          readableBuffer.addToBuffer(act);
          }
          break;
      }

      if (readBufferPos==readBuffer.length()) {
        if (reader!=null) {
	        reader.readEvent(this);
	      }
        sendWriteBuffer();
        processGo=false;
        readBuffer=new String();
        readBufferPos=0;
      }

    }
  }

  // ***************************************************************************************
  // MAIN TELNET PROTOCOL RESPONSES
  // ***************************************************************************************

  /** Main telnet protocol command management - WILL
      @param int option: the telnet option for that a WILL will be answered
  */
  private void replyWill(int option) {
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_WILL);
    addToWriteBuffer(option);
  }

  /** Main telnet protocol command management - WONT
      @param int option: the telnet option for that a WONT will be answered
  */
  private void replyWont(int option) {
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_WONT);
    addToWriteBuffer(option);
  }

  /** Main telnet protocol command management - DO
      @param int option: the telnet option for that a DO will be answered
  */
  private void replyDo(int option) {
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_DO);
    addToWriteBuffer(option);
  }

  /** Main telnet protocol command management - DONT
      @param int option: the telnet option for that a DONT will be answered
  */
  private void replyDont(int option) {
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_DONT);
    addToWriteBuffer(option);
  }

  // ***************************************************************************************
  // TELNET OPTION PROCESSING
  // ***************************************************************************************

  /** Main telnet protocol command management - Processing DO requests
      @param int option: the option for that a DO has been received
  */
  private void processDo(int option) {
    switch (option) {
      case MJC_TELOPT_TERMINAL_TYPE : // 24
        replyWill(option);
        break;
      case MJC_TELOPT_TERMINAL_SPEED : // 32
        replyWill(option);
        break;
      case MJC_TELOPT_XDISP_LOC : // 35
        replyWont(option);
        break;
      case MJC_TELOPT_NEW_ENVIRON : // 39
        replyWill(option);
        break;
      case MJC_TELOPT_ECHO : // 1
        replyWont(option);
        break;
      case MJC_TELOPT_NAWS : // 31
        replyWill(option);
        replyNaws();
        break;
      case MJC_TELOPT_TOGGLE_FLOW_CONTROL : // 33
        replyWont(option);
//        replyWill(option);
        break;
      default:
        replyWont(option);
        break;
    }
  }

  /** Main telnet protocol command management - Processing SB requests
      @param int option: the option for that an SB has been received
  */
  private void processSb(int option) {
    switch (option) {
      case MJC_TELOPT_TERMINAL_TYPE : // 24
        processSbTerminalType();
        break;
      case MJC_TELOPT_TERMINAL_SPEED : // 32
        processSbTerminalSpeed();
        break;
      case MJC_TELOPT_NEW_ENVIRON : //39
        processSbNewEnviron();
        break;
      case MJC_TELOPT_TOGGLE_FLOW_CONTROL : // 33
        processSbToggleFlowControl();
	      break;
      default:
//        replyDont(option);
//        replyWont(option);
        ts_unsupportedSb=true;
    }
  }

  /** Main telnet protocol command management - Processing WILL requests
      @param int option: the option for that a WILL has been received
  */
  private void processWill(int option) {
    switch (option) {
      case MJC_TELOPT_SUPPRESS_GO_AHEAD : // 3
        replyDo(option);
        break;
      case MJC_TELOPT_STATUS : // 5
        replyDo(option);
        break;
      case MJC_TELOPT_ECHO : // 1
        replyDo(option);
        break;
      default:
        replyDont(option);
        break;
    }
  }

  // ***************************************************************************************
  // TELNET OPTIONS
  // ***************************************************************************************

  /** Reply for Subnegotiation request for TERMINAL TYPE
  */
  private void processSbTerminalType() { // 24
    if (readBufferAvailable()>=3) {
      readBufferPos+=3; //1-??,255-IAC,240-SE
      addToWriteBuffer(MJC_TELOPT_IAC);
      addToWriteBuffer(MJC_TELOPT_SB);
      addToWriteBuffer(MJC_TELOPT_TERMINAL_TYPE);
      addToWriteBuffer(0);
      addToWriteBuffer(configuration.getString(CONF_TERMINAL_NAME, P_TERMINAL_NAME_DEF));
      addToWriteBuffer(MJC_TELOPT_IAC);
      addToWriteBuffer(MJC_TELOPT_SE);
    }
    else { // error !
    }
  }

  /** Reply for Subnegotionation request TOGGLE FLOW CONTROL
  */
  private void processSbToggleFlowControl() { // 24
    if (readBufferAvailable()>=3) {
      readBufferPos+=3; //1-??,255-IAC,240-SE
      // addToWriteBuffer(MJC_TELOPT_NOP, DP_COMMAND);
    }
    else { // error !
    }
  }

  /** Reply for Subnegotiation request for TERMINAL SPEED
  */
  private void processSbTerminalSpeed() { // 32
    if (readBufferAvailable()>=3) {
      readBufferPos+=3; //1-??,255-IAC,240-SE
      addToWriteBuffer(MJC_TELOPT_IAC);
      addToWriteBuffer(MJC_TELOPT_SB);
      addToWriteBuffer(MJC_TELOPT_TERMINAL_SPEED);
      addToWriteBuffer(0);
      addToWriteBuffer(configuration.getString(CONF_TERMINAL_SPEED_WRITE,
        P_TERMINAL_SPEED_DEF_WRITE)); // transmit rate
      addToWriteBuffer(","); // transmit rate
      addToWriteBuffer(configuration.getString(CONF_TERMINAL_SPEED_READ,
        P_TERMINAL_SPEED_DEF_READ)); // receive rate
      addToWriteBuffer(MJC_TELOPT_IAC);
      addToWriteBuffer(MJC_TELOPT_SE);
    }
    else { // error !
    }
  }

  /** Reply for Subnegotiation request for NEWENVIRON
  */
  private void processSbNewEnviron() { // 39
    if (readBufferAvailable()>=3) {
      readBufferPos+=3; //1-??,255-IAC,240-SE
      addToWriteBuffer(MJC_TELOPT_IAC);
      addToWriteBuffer(MJC_TELOPT_SB);
      addToWriteBuffer(MJC_TELOPT_NEW_ENVIRON);
      addToWriteBuffer(0);
      addToWriteBuffer(MJC_TELOPT_IAC);
      addToWriteBuffer(MJC_TELOPT_SE);
    }
    else { // error !
    }
  }

  /** Reply for subnegotiation request NAWS - 31
  */
  private void replyNaws() { // 31
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_SB);
    addToWriteBuffer(MJC_TELOPT_NAWS);
    addToWriteBuffer(0);
    addToWriteBuffer(configuration.getInt(CONF_DISPLAY_SIZE_WIDTH, P_DISPLAY_SIZE_DEF_WIDTH));
    addToWriteBuffer(0);
    addToWriteBuffer(configuration.getInt(CONF_DISPLAY_SIZE_HEIGHT, P_DISPLAY_SIZE_DEF_HEIGHT));
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_SE);
  }

  /** Reply for subnegotiation request TOGGLE FLOW CONTROL
  */
  private void replyToggleFlowControl() {
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_SB);
    addToWriteBuffer(MJC_TELOPT_TOGGLE_FLOW_CONTROL);
    addToWriteBuffer(2); // restart any
    addToWriteBuffer(MJC_TELOPT_IAC);
    addToWriteBuffer(MJC_TELOPT_SE);
  }

  // ***************************************************************************************
  // COMMUNICATION
  // ***************************************************************************************
  /** Send a string through the Telnet protocol
      @param String s: the String to be sent through the Telnet protocol
  */
  public void send(String s) {
    writeable.write(s);
    writeable.flush();
  }

  // ***************************************************************************************
  // CONFIGURATION
  // ***************************************************************************************

  /** Initalizes intern variables fromt he configuration.
  */
  private void initValuesFromConfig() {

    if (configuration==null) {
      readableBuffer=new MJCStringBuffer(P_READ_BUFFER_DEF_SIZE);
      writeBuffer=new char[P_WRITE_BUFFER_DEF_SIZE];
    }
    else {
      readableBuffer=new MJCStringBuffer(configuration.getInt(CONF_READ_BUFFER_SIZE,
        P_READ_BUFFER_DEF_SIZE));
      writeBuffer=new char[configuration.getInt(CONF_WRITE_BUFFER_SIZE,
        P_WRITE_BUFFER_DEF_SIZE)];
    }
    readBuffer=new String();
    writeBufferPos=0;
    readBufferPos=0;
  }

  /** Prints debug information to the standard output.
      @param String s: the debug information
  */
  private void debugOutput(String s) {
    System.out.println("[TP]"+s);
  }
}