/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.streams;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import java.io.*;

/** Class for creating a link between an InputStream and the MJC
    layer hierarchy. This class acts as a MJCReadable, this means it
    reads data from the specified InputStream and then forwards it
    to the MJCReader through the readEvent mechanism. Reading is
    buffered. Reading fromt the InputStream happens in a different
    thread as in which the MJCReadableHasAIS has been constructed.
*/    

public class MJCReadableHasAIS implements MJCReadable, Runnable {

  private static final String NAME="MJCReadableHasAIS";
  private static final String VERSION="0.1";
  private static final int bufSize=50000;
  
  private MJCReader reader;
  private MJCStringBuffer buf;
  private InputStream is;
  private boolean connected;
  
  /** Constructor. The InputStream to read from has to be specified.
      @param InputStream is: the InputStream to read from
  */  
  public MJCReadableHasAIS(InputStream is) {
    this.is=is;
    buf=new MJCStringBuffer(bufSize);
    connected=false;
  }
  
  // --------------------------------------------------------------------------
  // MJCReadable
  // --------------------------------------------------------------------------
  
  /** Returns the MJC name of the MJCReadable.
      @return String: the MJC name of the MJCReadable
  */      
  public String getReadableName() {
    return NAME;
  }
  
  /** Returns the MJC version of the MJCReadable.
      @return String: the MJC version of the MJCReadable
  */      
  public String getReadableVersion() {
    return VERSION;
  }
  
  /** Returns the number of bytes that can be read from this object.
      @return int: the number of bytes that can be read
  */      
  public int getAvailable() {
    return buf.available();    
  }
  
  /** Returns the contents of the read buffer.
      @return String: the contents of the read buffer
  */      
  public String getInput() {
    String temp=buf.emptyBuffer();
    synchronized (this) {
      notify();
    }
    return temp;  
  }
  
  /** Returns wether the InputStream is still connected or the
      stream has reached an end.
      @return boolean: wether the InputStream is still connected
  */      
  public boolean isReadableConnected() {
    return connected;
  }
  
  /** Sets the MJCReader to which the readEvents has to be sent.
      @param MJCReader reader: the MJCReader to which the readEvents
        have to be sent
  */	
  public void setReader(MJCReader reader) {
    this.reader=reader;
    connected=true;
  }
  
  /** Returns the MJCReader to which the readEvents will be sent.
      @return MJCReader: the MJCReader to which the readEvents will be 
        sent
  */	
  public MJCReader getReader() {
    return reader;
  }        
  
  // --------------------------------------------------------------------------
  // Runnable
  // --------------------------------------------------------------------------
  
  /** Entry point for the InputStream reader thread.
  */
  public void run() {
    listenForRead();
  }
  
  /** This method reads the InputStream and sends the readEvents to
      the MJCReader.
  */      
  private void listenForRead() {
    int firstByte;
    byte[] tempRead;
    String tempString;
    int t;
    
    try {
      while (reader==null) {
        synchronized (this) {
	  wait();
	}
      }	  
      while (true) {
        tempRead=new byte[1];
        firstByte=is.read();

        if (firstByte==-1) {
	  connected=false;
          return;
        }

        tempRead[0]=(byte)firstByte;
	while (!buf.addToBuffer(tempRead)) {
	  reader.readEvent(this);
	  synchronized (this) {
	    wait();
	  }
	}      
	
        while (is.available()>0) {
	  while (buf.available()==0) {
	    reader.readEvent(this);
	    synchronized (this) {
	      wait();
	    }
	  }      
          if (is.available()>buf.available()) {
	    tempRead=new byte[bufSize];
	  }
	  else {
	    tempRead=new byte[is.available()];
	  }
          t=is.read(tempRead);
          buf.addToBuffer(tempRead, t);
        }

        if (buf.getBufferLength()>0) {
          reader.readEvent(this);
        }

      }
    }
    catch (Exception e) {
      connected=false;
    }
  }
}