/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.streams;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import java.io.*;

/** Class for enabling to write to the MCJ layer hierarchy using the
    conventional OutputStream methods. Objects of this class act as
    OutputStreams, and they forward the data written on them 
    to MJCReaders.
*/    

public class MJCReadableIsAOS extends OutputStream implements MJCReadable {

  private static final String NAME="MJCReadableIsAOS";
  private static final String VERSION="0.1";
  private static final int bufSize=50000;
  
  private MJCStringBuffer buf;
  private boolean connected;
  private MJCReader reader;
  private boolean threadSuspended;
  
  /** Default constructor.
  */    
  public MJCReadableIsAOS() {
    connected=true;
    buf=new MJCStringBuffer(bufSize);
    threadSuspended=false;
  }
  
  // --------------------------------------------------------------------------
  // MJCReadable
  // --------------------------------------------------------------------------
  
  /** Returns the MJC name of the MJCReadable.
      @return String: the MJC name of the MJCReadable
  */  
  public String getReadableName() {
    return NAME;
  }
  
  /** Returns the MJC version of the MJCReadable.
      @return String: the MJC version of the MJCReadable
  */      
  public String getReadableVersion() {
    return VERSION;
  }
  
  /** Sets the MJCReader, which will query data from this object.
      @param MJCReader reader: the MJCReader that will query data
        from this object
  */	
  public void setReader(MJCReader reader) {
    this.reader=reader;
  }            
  
  /** Gets the MJCReader, which will query data from this object.
      @return MJCReader: the MJCReader that will query data from this
        object
  */	
  public MJCReader getReader() {
    return reader;
  }
  
  /** Returns the number of bytes that can be read from this MJCReader.
      @return int: the number of bytes that can be read
  */      
  public int getAvailable() {
    return buf.getBufferLength();
  }
  
  /** Returns the String that can be read from this MJCReader, this means
      the data that has been written to it as an OutputStream.
      @return String: the String that can be read
  */      
  public synchronized String getInput() {
    String temp=buf.emptyBuffer();
    threadSuspended=false;
    notify();
    return temp;
  }
  
  /** Returns wether the it could be read from this object.
      @return boolean: wether it could be read from this object
  */      
  public boolean isReadableConnected() {
    return connected;
  }    
  
  // --------------------------------------------------------------------------
  // OutputStream
  // --------------------------------------------------------------------------    
  
  /** This method closes the OuputStream. After it, nothing can
      be written on it. If there is still data in the buffer, 
      a readEvent is called for the MJCReader if specified.
      @exception IOException: if something goes wrong
  */      
  public void close() throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (buf.getBufferLength()>0) {
      if (reader==null) throw new IOException("MJCReader not specified.");
      reader.readEvent(this);
    }
    connected=false;  
  }
  
  /** Flushes the read buffer to the MJCReader. As reading in the
      MJC layer hierarchy is asynchronous, the current thread will
      be suspended after a readEvent is called at the MJCReader till
      the getInput method of this object is called.
      @exception IOException: if something goes wrong
  */      
  public void flush() throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (buf.getBufferLength()>0) {
      if (reader==null) throw new IOException("MJCReader not specified.");
      threadSuspended=true;
      reader.readEvent(this);
      synchronized (this) {              
	try {
	  if (threadSuspended) {
    	    wait();
	  }    
	}
	catch (Exception e) {
	}  
      }	
    }
  }
  
  /** Writes an int (0...255) to the MJCReader. Writing is buffered.
      @param int b: the int to be written
      @exception IOException: if something goes wrong
  */      
  public void write(int b) throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (reader==null) throw new IOException("MJCReader not specified.");
    while (!buf.addToBuffer((char)b)) {      
      threadSuspended=true;
      reader.readEvent(this);
      synchronized (this) {        
	try {
	  if (threadSuspended) {
            wait();
	  }    
	}
	catch (Exception e) {
	}
      }
    }  	
  }
  
  /** Writes a byte array to the MJCReader. Writing is buffered.
      @param byte[] b: the byte array to be written
      @exception IOException: if something goes wrong
  */      
  public void write(byte b[]) throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (reader==null) throw new IOException("MJCReader not specified.");
    int sP=0;
    int bL=0;
    boolean doLoop=true;
    while (doLoop) {
      bL=buf.available();
      if ((b.length-sP)>bL) {
        buf.addToBuffer(b,sP,bL);
	sP+=bL;
	threadSuspended=true;
	reader.readEvent(this);
	synchronized (this) {          
	  try {	  
	    if (threadSuspended) {
	      wait();
	    }
	  }
	  catch (Exception e) {
	  }
	}
	if (sP==b.length) {
	  doLoop=false;
	}  
      }
      else {
        buf.addToBuffer(b,sP,b.length-sP);
	doLoop=false;
      }
    }
  }
  
  /** Writes a part (specified with start offset and length) of a
      byte array to the MJCReader. Writing is buffered.      
      @param byte[] b: the byte array to be written
      @param int off: the starting offset of the part
      @param int len: the length of the part
      @exception IOException: if something goes wrong
  */      
  public void write(byte b[], int off, int len) throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (reader==null) throw new IOException("MJCReader not specified.");
    int sP=off;
    int bL=0;
    boolean doLoop=true;
    while (doLoop) {
      bL=buf.available();
      if (len>bL) {
        buf.addToBuffer(b,sP,bL);
	sP+=bL;
	len-=bL;
	threadSuspended=true;
	reader.readEvent(this);
	synchronized (this) {	          
	  try {
	    if (threadSuspended) {
       	      wait();
	    }
	  }
	  catch (Exception e) {
	  }    
	}
	if (len==0) {
	  doLoop=false;
	}  
      }
      else {
        buf.addToBuffer(b,sP,len);
	doLoop=false;
      }
    }
  }
  
}