/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.streams;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import java.io.*;

/** Class for creating a link between an OutputStream and the MJC layer
    hierarchy. This class acts as a MJCWriteable, this means an upper
    layer can write to it, and what has been written on it, will be written
    again to the OutputStream given. The writing to the OutputStream is 
    buffered.
*/

public class MJCWriteableHasAOS implements MJCWriteable {
  
  private static final String NAME="MJCWriteableHasAOS";
  private static final String VERSION="0.1";
  private static final int bufSize=50000;
  
  private OutputStream os;
  private MJCStringBuffer buf;
  private String lastError;
  
  private boolean connected;
  
  /** Constructor.
      @param OutputStream os: the specified OutputStream to which the data has to
        be written
  */	  
  public MJCWriteableHasAOS(OutputStream os) {
    this.os=os;
    buf=new MJCStringBuffer(bufSize);
    connected=true;
    lastError=new String();
  }
  
  /** Returns the MJC name of the MJCWriteable.
      @return String: the MJC name of the MJCWriteable
  */      
  public String getWriteableName() {
    return NAME;
  }
  
  /** Returns the MJC version of the MJCWriteable.
      @return String: the MJC version of the MJCWriteable
  */      
  public String getWriteableVersion() {
    return VERSION;
  } 
  
  /** Writes a single char (0...255) to the OutputStream. Writing to the
      OutputStream is buffered.
      @param char c: the char to be written to the OutputStream
      @return boolean: wether the writing has been successful
  */      
  public boolean write(char c) {
    if (connected) {
      if (!buf.addToBuffer(c)) {
        flush();
        return buf.addToBuffer(c);  
      }
      else {
        return true;	
      }	
    }
    else {
      return false;
    }    
  }
  
  /** Writes a String to the OutputStream. Writing to the
      OutputStream is buffered.
      @param String s: the String to be written to the OutputStream
      @return boolean: wether the writing has been successful
  */
  public boolean write(String s) {
    if (connected) {
      if (!buf.addToBuffer(s)) {
         flush();
         if (!buf.addToBuffer(s)) {
           writeOut(s);
	 }  
      }
      return true;    
    }
    else {
      return false;
    }    
  }
  
  /** Writes an array of chars to the OutputStream. Writing to the
      OutputStream is buffered.
      @param char[] c: the char array to be written to the OutputStream
      @return boolean: wether the writing has been successful
  */
  public boolean write(char c[]) {
    if (connected) {
      if (!buf.addToBuffer(c)) {
        flush();
        if (!buf.addToBuffer(c)) {
          writeOut(new String(c));
        }
      }	
      return true;    
    }  
    else {
      return false;
    }  
  }
  
  /** Flushes the conents of the buffer to the OutputStream.
  */
  public void flush() {
    if (connected) {
      writeOut(buf.emptyBuffer());
    }  
  }
  
  /** Flushed the contents of the buffer to the OutputStream.
      @return boolean: wether the flushing has been successful
  /*
  private boolean flushB() {
    if (connected) {
      return writeOut(buf.emptyBuffer());
    }
    else {
      return false;
    }
  }            
  
  /** Writes a String directly to the OutputStream. This method does not
      use the buffer.
      @param String s: the String to be written directly to the OutputStream
      @return boolean: wether the writing to the OutputStream has been successful
  */      
  private boolean writeOut(String s) {
    try {
      for (int i=0; i<s.length(); i++) {
        os.write((byte)s.charAt(i));
      }
      os.flush();
      return true;
    }
    catch (IOException e) {
      connected=false;
      lastError=e.getMessage();
    }    
    return false;
  }         
  
  /** Returns the description of the last occured error. Once queried, the
      errormessage will be deleted.
      @return String: the error description of the error that occured 
        since the last query
  */	
  public String getLastWriteableError() {
    String temp=lastError;
    lastError=new String();
    return temp;
  }
  
  /** Returns wether this MJCWriteable object is still connected to the
      OutputStream, that means, wether this object can still write to the
      OutputStream.
      @return boolean: wether this object is connected to its OutputStream
  */      
  public boolean isWriteableConnected() {
    return connected;
  }      
}