/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.streams;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import java.io.*;

/** Class for enabling to read from the MJC layer hierarchy using the
    conventional InputStream methods. Objects of this class act as 
    InputStreams, the data readable from the InputStream will be
    generated by the MJCWriter corresponding to this MJCWriteable.
*/

public class MJCWriteableIsAIS extends InputStream implements MJCWriteable {

  private static final String NAME="MJCWriteableIsAIS";
  private static final String VERSION="0.1";
  private static final int bufSize=50000;
  
  private MJCStringBuffer buf;
  private boolean isClosed;
  private boolean threadSuspendedW;
  private boolean threadSuspendedR;
  private String lastError;
  
  /** Default constructor.
  */  
  public MJCWriteableIsAIS() {
    buf=new MJCStringBuffer(bufSize);
    isClosed=false;
    threadSuspendedW=false;
    threadSuspendedR=false;
    lastError=new String();
  }
  
  // --------------------------------------------------------------------------
  // MJCWriteable
  // --------------------------------------------------------------------------
  
  /** Returns the MJC name of the MJCWriteable.
      @return String: the MJC name of the MJCWriteable
  */      
  public String getWriteableName() {
    return NAME;
  }
  
  /** Returns the MJC version of the MJCWriteable.
      @return String: te MJC version of the MJCWriteable
  */        
  public String getWriteableVersion() {
    return VERSION;
  }    
  
  /** Writes one single char to the stream. Writing is buffered.
      @param char c: the char to be written to the stream
      @return boolean: wether the writing has been successful
  */      
  public boolean write(char c) {    
    if (isClosed) {
      lastError="InputSream has been closed.";
      return false;
    }  
    boolean doLoop=true;
    while (doLoop) {
      if (buf.available()>0) {
        buf.addToBuffer(c);
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	}  	
	doLoop=false;
      }
      else {	
	threadSuspendedW=true;
	synchronized (this) {
	  try {
	    if (threadSuspendedW) {
  	      wait();
	    }  
	  }
	  catch (Exception e) {
	  }
	}
      }	
    }  	
    return true;
  }

  /** Writes a String to the stream. Wrinting is buffered.
      @param String s: the String to be written to the stream
      @return boolean: wether the writing has been successful
  */        
  public boolean write(String s) {
    if (isClosed) {
      lastError="InputSream has been closed.";
      return false;
    }  
    int bL=0;
    boolean doLoop=true;
    int sP=0;
    while (doLoop) {
      bL=buf.available();
      if ((s.length()-sP)>bL) {
        buf.addToBuffer(s,sP,bL);
	sP+=bL;
	threadSuspendedW=true;
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	  try {
	    if (threadSuspendedW) {
  	      wait();
	    }  
	  }
	  catch (Exception e) {
	  }
	}
	if (sP==s.length()) {
	  doLoop=false;
	}
      }	
      else {
        buf.addToBuffer(s,sP,s.length()-sP);
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	}  
	doLoop=false;
      }
    }  	
    return true;
  }
  
  /** Writes a char array to the stream. Writing is buffered.
      @param char[] c: the char array to be written to the stream
      @return boolean: wether the writing has been successful
  */            
  public boolean write(char[] c) {
    if (isClosed) {
      lastError="InputSream has been closed.";
      return false;
    }  
    int bL=0;
    boolean doLoop=true;
    int sP=0;
    while (doLoop) {
      bL=buf.available();
      if ((c.length-sP)>bL) {
        buf.addToBuffer(c,sP,bL);
	sP+=bL;
	threadSuspendedW=true;
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	  try {
	    if (threadSuspendedW) {
  	      wait();
	    }  
	  }
	  catch (Exception e) {
	  }
	}
	if (sP==c.length) {
	  doLoop=false;
	}
      }	
      else {
        buf.addToBuffer(c,sP,c.length-sP);
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	}  
	doLoop=false;
      }
    }  	
    return true;
  }
  
  /** MJCWriteableIsAIS does not implement the flush method, because
      it cannot be ensured that after flush() the contents of the
      buffer will be read from the InputStream. This method does 
      nothing.
  */      
  public void flush() {
  }
  
  /** Returns the last error that occured during the write methods. 
      After this method has been invoded, the errormessage will
      be erased.
      @return String: the last error message
  */      
  public String getLastWriteableError() {
    String temp=lastError;
    lastError=new String();
    return temp;
  }
  
  /** Returns wether the InputStream is still open.
      @return boolean: wether the InputStream is still open
  */      
  public boolean isWriteableConnected() {
    return (!isClosed);
  }
  
  // --------------------------------------------------------------------------
  // InputStream
   // --------------------------------------------------------------------------

  /** Reads 1 single byte from the stream. The read blocks till data
      is available.
      @return int: the byte read
      @exception IOException: if the InputStream has been closed
  */      
  public synchronized int read() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    
    while (buf.getBufferLength()==0) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }  	  
    
    String temp=buf.emptyBuffer(1);
    threadSuspendedW=false;
    notify();
    return ((int)temp.charAt(0));      
  }
  
  /** Reads b.length bytes from the stream. The read blocks till data
      is available. Returns the number of bytes read.
      @param byte[] b: the byte array into the data will be read
      @return int: the number of bytes read
      @exception IOException: if the InputStream has been closed
  */      
  public synchronized int read(byte[] b) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    
    String temp;
    
    int len;
    if (b.length>bufSize) len=bufSize; else len=b.length;
    
    while (buf.getBufferLength()<len) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }
    
    temp=buf.emptyBuffer(len);
    threadSuspendedW=false;
    notify();
    for (int i=0; i<temp.length(); i++) b[i]=(byte)temp.charAt(i);
    return temp.length();
  }
        
  /** Reads len bytes from the stream. The read blocks till data
      is available. Returns the number of bytes read.
      @param byte[] b: the byte array into the data will be read
      @param int off: the beginning offset from which the bytes will
        be written into b (inclusive)
      @param int len: the number of bytes to be read	
      @return int: the number of bytes read
      @exception IOException: if the InputStream has been closed
      @exception IndexOutOfBoundsException: if (b.length<(off+len))
  */      
  public synchronized int read(byte[] b, int off, int len) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    
    String temp;
    
    int lenR;
    if (b.length>bufSize) lenR=bufSize; else lenR=b.length;
    
    while (buf.getBufferLength()<lenR) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }
    
    temp=buf.emptyBuffer(len);
    threadSuspendedW=false;
    notify();
    for (int i=0; i<temp.length(); i++) b[off+i]=(byte)temp.charAt(i);
    return temp.length();
  }
  
  /** Skips n bytes in the stream. Skipped data will be lost. For
      some reasons the stream may skip less bytes than n.
      @param long n: number of bytes to be skipped
      @return long: number of bytes actually skipped
      @exception IOException: if the InputStream has been closed
  */      
  public long skip(long n) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
  
    int ret;
    if (n>buf.getBufferLength()) {
      ret=buf.getBufferLength();
      buf.emptyBuffer();
    }
    else {
      ret=(int)n;
      buf.emptyBuffer((int)n);  
    }
    synchronized (this) {
      threadSuspendedW=false;
      notify();
    }
    return (long)ret;
  }        
  
  /** Returns the number of bytes that are currently available
      for reading from the stream.
      @return int: the number of bytes available from the stream
      @exception IOException: if the InputStream has been closed
  */      
  public int available() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");  
    return (buf.getBufferLength());
  }
  
  /** Closes the InputStream. After close(), no read, close and available
      methods can be called without IOException.
      @exception IOException: if the InputStream has been closed
  */      
  public void close() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");  
    isClosed=true;
  }
  
  /** MJCWriteableIsAIS does not support mark/reset. This method
      does actually nothing.
  */      
  public synchronized void mark(int readlimit) {
  }
  
  /** MJCWriteableIsAIS does not support mark/reset. This method will
      always throw an IOException.
      @exception IOException: MJCWriteableIsAIS does not support mark/reset
  */      
  public synchronized void reset() throws IOException {
    throw new IOException("Mark not supported.");
  }
  
  /** MJCWriteableIsAIS does not support mark/reset.
      @return boolean: false
  */      
  public boolean markSupported() {
    return false;
  } 
   
}