/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.streams;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import java.io.*;

/** Class for creating a link between an InputStream and the MJC
    layer hierarchy. This class acts as a MJCWriter, this means it writes 
    data read from the specified InputStream will be written to 
    the MJCWriteable. Writing is not buffered. Reading from the 
    InputStream happens in a different thread as in which the 
    MJCWriterHasAIS has been constructed.    
*/    

public class MJCWriterHasAIS implements MJCWriter, Runnable {

  private static final String NAME="MJCWriterHasAIS";
  private static final String VERSION="0.1";
  private static final int maxLength=50000;
  
  private MJCWriteable writeable;
  private InputStream is;
  private Thread myThread;
  
  /** Constructor. The InputStream to read from has to be specified. 
      @param InputStream is: the InputStream to read from
  */           
  public MJCWriterHasAIS(InputStream is) {
    this.is=is;
    myThread=new Thread(this);
    myThread.start();
  }

  // --------------------------------------------------------------------------
  // MJCReadable
  // --------------------------------------------------------------------------
  
  /** Returns the MJC name of the MJCWriter.
      @return String: the MJC name of the MJC writer
  */      
  public String getWriterName() {
    return NAME;
  }
  
  /** Returns the MJC version of the MJCWriter.
      @return String: the MJC version of the MJCWriter
  */      
  public String getWriterVersion() {
    return VERSION;
  }
  
  /** Sets the MJCWriteable to which data read from the InputStream
      will be written.
      @param MJCWriteable writeable: the MJCWriteable to which data
        read from the InputStram will be written
  */	
  public void setWriteable(MJCWriteable writeable) {
    this.writeable=writeable;
    synchronized (this) {
      notify();
    }  
  }
  
  /** Returns the MJCWriteable to which data read from the InputStream
      will be written.
      @return MJCWriteable: the MJCWriteable to which data read from the
        InputStream will be written
  */	
  public MJCWriteable getWriteable() {
    return writeable;
  }                  
  
  // --------------------------------------------------------------------------
  // Runnable
  // --------------------------------------------------------------------------
  
  /** The entry point for the InputStream reader thread.
  */
  public void run() {
    listenForRead();
  }
  
  /** This method reads the InputStream and writes the data read
      from it to the MCJWriteable.
  */      
  private void listenForRead() {
    int firstByte;
    byte[] tempRead;
    String tempString;
    int t;

    try {
      while (writeable==null) {
        synchronized (this) {
          wait();
        }  
      }
    
      while (true) {
        tempRead=new byte[1];
        firstByte=is.read();

        if (firstByte==-1) {
          return;
        }
	      
        if (!writeable.write((char)firstByte)) {
	  return;
	}  
	  
        while (is.available()>0) {
	  if (is.available()>maxLength) {
	    tempRead=new byte[maxLength];
	  }    
	  else {    
            tempRead=new byte[is.available()];
	  }    
          t=is.read(tempRead);
	  tempString=new String(tempRead,0,t);
	  if (!writeable.write(tempString)) {
	    return;
	  }    	  
	  writeable.flush();
         }        
      }
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }
        
}