/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.streams;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import java.io.*;

/** Class to enable to write into the MJC layer hierarchy using the
    conventional OutputStream methods. Objects of this class act as
    OutputStreams, and they forward the data written on them 
    to MJCWriteables.
*/

public class MJCWriterIsAOS extends OutputStream implements MJCWriter {

  private static final String NAME="MJCWriterIsAOS";
  private static final String VERSION="0.1";
  private static final int bufSize=50000;
  
  private MJCStringBuffer buf;
  private MJCWriteable writeable;
  private boolean connected;
  
  /** Default constructor.
  */
  public MJCWriterIsAOS() {
    connected=true;
  }    
  
  // ------------------------------------------------------------------------------------
  // MJCWriter 
  // ------------------------------------------------------------------------------------

  /** Returns the MJC name of the MJCWriter.
      @return String: the MJC name of the MJCWriter
  */      
  public String getWriterName() {
    return NAME;
  }
  
  /** Returns the MJC version of the MJCWriter.
      @return String: the MJC version of the MJCWriter
  */      
  public String getWriterVersion() {
    return VERSION;
  }    
  
  /** Sets the MJCWriteable to which this object writes.
      @param MJCWriteable writeable: the MCJWriteable to which this object writes
  */      
  public void setWriteable(MJCWriteable writeable) {
    this.writeable=writeable;
    connected=true;
    buf=new MJCStringBuffer(bufSize);
  }
  
  /** Returns the MJCWriteable to which this object writes.
      @return MJCWriteable: the MJCWriteable to which this object writes
  */      
  public MJCWriteable getWriteable() {    
    return writeable;
  }    
  
  // ------------------------------------------------------------------------------------
  // OutputStream
  // ------------------------------------------------------------------------------------
  
  /** Writes an int (0...255) to the MJCWriteable of this object. Writing is buffered.
      @param int b: the int to be written to the MCJWriteable
      @exception IOException: if something goes wrong
  */      
  public void write(int b) throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("MJCWriteable not specified.");
    if (!buf.addToBuffer((char)b)) {
      flush();
      if (!buf.addToBuffer((char)b)) {
        connected=false;
        throw new IOException("Buffer too small.");	
      }	
    }
  }
  
  /** Writes a byte array to the MJCWriteable of this object. Writing is buffered.
      @param byte[] b: the byte array to be written to the MCJWriteable
      @exception IOException: if something goes wrong
  */
  public void write(byte b[]) throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("MJCWriteable not specified.");
    if (!buf.addToBuffer(b)) {
      flush();
      if (!buf.addToBuffer(b)) {
        writeOut(new String(b));
      }
    }  
  }
  
  /** Writes a part (specified with beginning offset and length) of a byte array 
      to the MJCWriteable of this object. Writing is buffered.
      @param byte[] b: the byte array to be written to the MCJWriteable
      @param int off: the beginning offset (inclusive) of the part to be written out
      @param int len: the length of the part to be written out
      @exception IOException: if something goes wrong
  */
  public void write(byte[] b, int off, int len) throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("MJCWriteable not specified.");
    if (!buf.addToBuffer(b, off, len)) {
      flush();
      if (!buf.addToBuffer(b, off, len)) {
        writeOut(new String(b, off, len));
      }
    }  
  }
  
  /** Flushes the contents of the buffer to the MJCWriteable of this object.
      @throw IOException: if something goes wrong
  */      
  public void flush() throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("MJCWriteable not specified.");    
    writeOut(buf.emptyBuffer());
  }
  
  /** Closes the OuputStream. After a successful close, nothing can be
      written to it.
      @exception IOException: if something goes wron
  */      
  public void close() throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (buf.getBufferLength()>0) {
      flush();
    }  
    connected=false;
  }
 
  /** Writes a String directly to the MJCWriteable of this object. Writing is
      not buffered.
      @exception IOException: if something goes wrong
  */      
  private void writeOut(String s) throws IOException {
    if (writeable==null) throw new IOException("MJCWriteable not specified.");
    if (!writeable.write(s)) {
      throw new IOException("Writeable returned error string: "+
        writeable.getLastWriteableError());
    }	
  }
}