/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.transport;

import java.net.*;
import java.io.*;
import java.util.*;
import com.monica.javacom.interfaces.*;
import com.monica.javacom.transport.*;
import com.monica.javacom.util.*;

/** This class provides TCP based connection over IP as a MJCTransport. MJCTCPTransport
    provides both server and client services. This class implements buffering in both
    directions.<br><br>
    Configureable values:
    <ul>
      <li><b>transportReadBufferSize</b>: the maximum size of the read buffer
      <li><b>transportWriteBufferSize</b>: the maximum size of the write buffer
    </ul>
*/
public class MJCTCPTransport implements MJCTransport {

  /* ---------------------------------------------------------------------------------- */
  /* Variables */
  /* ---------------------------------------------------------------------------------- */

  // Socket communication
  private ServerSocket server;
  private Socket clientOfServer;
  private Socket serverOfClient;

  // Network information
  private int port;
  private String host;

  // Current status
  private int status;
  private MJCConfiguration configuration;
  private String lastError;

  // Thread handling
  private Vector childrenClients;
  private MJCTransport ownerComm;

  // Connection streams
  private OutputStream output;
  private InputStream input;

  // IO
  private MJCStringBuffer readBuffer;
  private MJCStringBuffer writeBuffer;
  private MJCConverter converter;

  // MJCReadable
  private MJCReader reader;
  private MJCServerManager serverManager;

  private static final String CONF_READ_BUFFER_SIZE="transportReadBufferSize";
  private static final String CONF_WRITE_BUFFER_SIZE="transportWriteBufferSize";

  private static final int CONF_DEF_READ_BUFFER_SIZE=50000;
  private static final int CONF_DEF_WRITE_BUFFER_SIZE=50000;

  /* ---------------------------------------------------------------------------------- */
  /* Constants */
  /* ---------------------------------------------------------------------------------- */

  private static final String NAME="MJCTCPTransport";
  private static final String VERSION="0.1";

  /* ---------------------------------------------------------------------------------- */
  /* Constructors */
  /* ---------------------------------------------------------------------------------- */

  /** Normal constructor. Should be called upon creation.
  */
  public MJCTCPTransport() {
    status=MJCTransportConstants.NOT_CONNECTED;
    childrenClients=new Vector();
    initConfiguration();
    lastError=new String();
    converter=new MJCConverter();
  }

  /** Create a new thread for a new server.
      @param Socket client: the client of this server thread
  */
  public MJCTCPTransport(Socket client) throws IOException {
    this.clientOfServer=client;
    status=MJCTransportConstants.CONNECTED_SERVER;
    initConfiguration();
    initStreams();
    lastError=new String();
    converter=new MJCConverter();
  }

  /* ---------------------------------------------------------------------------------- */
  /* Initializing */
  /* ---------------------------------------------------------------------------------- */

  /** Initializes the MJCTransport as the server side. Clients
      can connect to the specified port. To every client a new server thread will
      be started.
      @param int port: the port where the clients can connect
      @exception IOException: if the server could not be started
  */
  public void initAsServer(int port) throws IOException {
    if (status==MJCTransportConstants.NOT_CONNECTED) { // not connected
      this.port=port;
      server=new ServerSocket(port);
      status=MJCTransportConstants.ACCEPTING_SOCKETS;
      while (true) {
        clientOfServer=server.accept();
        MJCTCPTransport m=new MJCTCPTransport(clientOfServer);
        m.setConfiguration(configuration);
        clientOfServer=null;
        Thread t=new Thread(m);
        childrenClients.addElement(m);
        serverManager.newClientHasArrivedEvent(this, m);
        //	debugOutput("Serverthread starting");
        t.start();
      }
    }
  }

  /** Initializes the MJCTransport as the client side. This client
      will try to connect to the specified host and port. The client reads
      from the server on  new thread.
      @param String host: the destination host
      @param int port: the destination port
      @exception IOException: if the client cannot connect to
        the specified host and port
  */
  public void initAsClient(String host, int port) throws IOException {
    if (status==MJCTransportConstants.NOT_CONNECTED) { // not connected
      this.port=port;
      this.host=host;
      serverOfClient=new Socket(host,port);
      Thread clientThread=new Thread(this);
      status=MJCTransportConstants.CONNECTED_CLIENT;
      initStreams();
      clientThread.start();
    }
  }

  /** Runnable run method. Entry point for the read thread.
  */
  public void run() {
    // debugOutput("thread started.");
    int locstat=status;
    listenForRead();
   // debugOutput("TCP Read Thread stopped. = "+locstat);
  }

  /** Method that waits for byte available to be read out from the socket. If those
      arrive, reads them out, sends event to owner and waits again. While waiting
      thread blocks.
  */
  void listenForRead() {
    int firstByte;
    byte[] tempRead;
    int t;

    try {
      while (true) {
        tempRead=new byte[1];
        firstByte=input.read();

        if (firstByte==-1) {
          closeConnections();
          return;
        }

        tempRead[0]=(byte)firstByte;
        addToReadBuffer(tempRead,1);

        while (input.available()>0) {
          if (input.available()>readBuffer.getMaxBufferLength()) {
            tempRead=new byte[readBuffer.getMaxBufferLength()];
          }
          else {
            tempRead=new byte[input.available()];
          }
          t=input.read(tempRead);
          addToReadBuffer(tempRead, t);
        }

        if (readBuffer.getBufferLength()>0) {
	        Thread.currentThread().sleep(10);
          reader.readEvent(this);
        }

      }
    }
    catch (Exception e) {
      lastError=e.toString();
      e.printStackTrace();
      closeConnections();
    }
  }

  /** Adds a byte array with the specified length to the read buffer.
      @param byte[] b: the byte array to be added to the read buffer
      @param int l: the length of the byte array
  */
  private void addToReadBuffer(byte[] c, int l) {
    if (!readBuffer.addToBuffer(c, l)) {
      synchronized (this) {
 	      try {
  	      Thread.currentThread().sleep(1);
	      }
 	      catch (Exception e) {
	      }
        reader.readEvent(this);
        while (!readBuffer.addToBuffer(c, l)) {
          try {
            synchronized (this) {
              wait();
            }
          }
          catch (Exception e) {
          }
        }
      }
    }
    else {
    }
  }

  /** Adds a String to the write buffer.
      @param String s: the String to be added to the write buffer.
  */
  private void addToWriteBuffer(String s) {
    // debugOutput("tr.atwb");
    if (isWriteableConnected()) {
      if (!writeBuffer.addToBuffer(s)) {
        flush();
        if (!writeBuffer.addToBuffer(s)) {
          writeOut(s);
        }
      }
    }
  }


  /** Flushes the contents of the write buffer to bottom layers.
  */
  public void flush() {
    // debugOutput("tr.fl");
    if (isWriteableConnected()) {
      writeOut(writeBuffer.emptyBuffer());
    }
  }

  /** Writes out a String directly to the TCP socket without buffering.
      @param String s: the String to be written out
  */
  public void writeOut(String s) {
    //  debugOutput("tr.wo+"+isWriteableConnected()+"; "+s.length());

    if (isWriteableConnected()) {
      // for (int i=0; i<s.length(); i++) debugOutput("W "+(int)s.charAt(i)+" : "+s.charAt(i));

  		try {
        output.write(converter.getBytes(s));
        output.flush();
      }
		  catch (Exception e) {
		  }
    }
    else {
      // debugOutput("writeable not connected.");
    }
  }

  /** Get back the data arrived from the socket and read out by listenForRead.
      @return String s: the contents of the read buffer
  */
  public synchronized String getInput() {
    // debugOutput("gi");
    String temp;
    if ((isReadableConnected()) || (status==MJCTransportConstants.NOT_CONNECTED_WITH_DATA_WAITING)) {
      temp=readBuffer.emptyBuffer();
      notify();
      return temp;
    }
    else {
      return new String();
    }
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCTransport */
  /* ---------------------------------------------------------------------------------- */

  /** Close all active connections
  */
  public void closeConnections() {
//    debugOutput("cl");
    int i;

    switch (status) {
      case MJCTransportConstants.NOT_CONNECTED :
        break;

      case MJCTransportConstants.CONNECTED_CLIENT :
        if (readBuffer.getBufferLength()>0) {
          status=MJCTransportConstants.NOT_CONNECTED_WITH_DATA_WAITING;
        }
        else {
          status=MJCTransportConstants.NOT_CONNECTED;
        }
        try {
          serverOfClient.close();
        }
        catch (Exception e) {
        }
        if (reader!=null) {
          reader.connectionClosedEvent(this);
        }
        break;

      case MJCTransportConstants.CONNECTED_SERVER :
        if (readBuffer.getBufferLength()>0) {
          status=MJCTransportConstants.NOT_CONNECTED_WITH_DATA_WAITING;
        }
        else {
          status=MJCTransportConstants.NOT_CONNECTED;
        }
        try {
          clientOfServer.close();
        }
        catch (Exception e) {
        }
        toDefaultSettings();
        if (reader!=null) {
          reader.connectionClosedEvent(this);
        }
        break;

      case MJCTransportConstants.ACCEPTING_SOCKETS :
        status=MJCTransportConstants.NOT_CONNECTED;
        for (i=0; i<childrenClients.size(); i++) {
          ((MJCTransport)(childrenClients.elementAt(i))).closeConnections();
        }
        try {
          server.close();
        }
        catch (Exception e) {
        }
        break;

      case MJCTransportConstants.NOT_CONNECTED_WITH_DATA_WAITING :
        break;
    }

    toDefaultSettings();
  }

  /** Returns the MJC name of the MJCTransport.
      @return String: the MJC name of the MJCTransport
  */
  public String getTransportName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCTransport.
      @return String: the MJC version of the MJCTransport
  */
  public String getTransportVersion() {
    return VERSION;
  }

  /** Sets the MJCReader to which the readEvents have to be sent.
      @param MJCReader reader: the MJCReader to which the readEvents
        have to be sent
  */
  public void setReader(MJCReader reader) {
    this.reader=reader;
  }

  /** Returns the MJCReader to which readEvents will be sent.
      @return MJCReader: the MJCReader to which readEvents will be sent
  */
  public MJCReader getReader() {
    return reader;
  }

  /** Sets the MJCServerManager which will be notified if a client
      connects.
      @param MJCServerManager serverManager: the MJCServerManager
        that has to be notified if a client connects
  */
  public void setServerManager(MJCServerManager serverManager) {
    this.serverManager=serverManager;
  }

  /** Returns the MJCServerManager which will be notified if a client
      connects.
      @return MJCServerManager: the MJCServerManager which will
        be notified if a client connects
  */
  public MJCServerManager getServerManager() {
    return serverManager;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCConfigureable */
  /* ---------------------------------------------------------------------------------- */

  /** Sets the MJCConfiguration for this object.
      @param MJCConfiguration configuration: the MJCConfiguration
        object for the customization
  */
  public void setConfiguration(MJCConfiguration configuration) {
    this.configuration=configuration;
    initConfiguration();
  }

  /** Returns the MJCConfiguration with which this object is customized.
      @return MJCConfiguration: the MJCConfiguration with which this
        object is customized
  */
  public MJCConfiguration getConfiguration() {
    return configuration;
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCWriteable */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCWriteable.
      @return String: the MJC name of the MJCWriteable
  */
  public String getWriteableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCWriteable.
      @return String: the MJC version of the MJCWriteable
  */
  public String getWriteableVersion() {
    return VERSION;
  }

  /** Writes one single char. Writing is synchronous and can be
      buffered.
      @param char c: the char to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(char c) {
    char[] charArray;
    if (isWriteableConnected()) {
      charArray=new char[1];
      charArray[0]=c;
      addToWriteBuffer(new String(charArray));
      return true;
    }
    return false;
  }

  /** Writes a String. Writing is synchronous and can be buffered.
      @param String s: the String to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(String s) {
//    debugOutput("tr.wS");
    if (isWriteableConnected()) {
      addToWriteBuffer(s);
      return true;
    }
    return false;
  }

  /** Writes a char array. Writing is synchronous and can be buffered.
      @param char[] c: the char array to be written
      @return boolean: wether the writing has been successful
  */
  public boolean write(char[] c) {
    if (isWriteableConnected()) {
      addToWriteBuffer(new String(c));
      return true;
    }
    return false;
  }

  /** Returns the error message of the last occured error. After
      returning it, the error message will be deleted and cannot
      be queried again.
      @return String: the last error message
  */
  public String getLastWriteableError() {
    String temp=lastError;
    lastError=new String();
    return temp;
  }

  /** Returns wether the MJCWriteable is connected to its target.
      @return boolean: wether the MJCWriteable is connected to
        its target
  */
  public boolean isWriteableConnected() {
    switch (status) {
      case MJCTransportConstants.CONNECTED_SERVER :
        return true;
      case MJCTransportConstants.CONNECTED_CLIENT :
        return true;
      default:
        return false;
    }
  }

  /* ---------------------------------------------------------------------------------- */
  /* MJCReadable */
  /* ---------------------------------------------------------------------------------- */

  /** Returns the MJC name of the MJCReadable.
      @return String: the MJC name of the MJCReadable
  */
  public String getReadableName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCReadable.
      @return String: the MJC version of the MJCReadable
  */
  public String getReadableVersion() {
    return VERSION;
  }

  /** Returns the number of bytes that are available for read out from
      this MJCReadable.
      @return int: the number of bytes available for read out
  */
  public int getAvailable() {
    return readBuffer.getBufferLength();
  }

  /** Returns wether the MJCReadable is still connected to its source.
      @return boolean: wether the MJCReadable is still connected to its
        source
  */
  public boolean isReadableConnected() {
    switch (status) {
      case MJCTransportConstants.CONNECTED_CLIENT :
        return true;
      case MJCTransportConstants.CONNECTED_SERVER :
        return true;
      default:
        return false;
    }
  }

  /** Sets class intern variables to their default settings.
  */
  private void toDefaultSettings() {
    server=null;
    clientOfServer=null;
    serverOfClient=null;
    childrenClients=null;
    input=null;
    output=null;
  }

  /**  Initialize stream variables.
  */
  private void initStreams() throws IOException {
    switch (status) {
      case MJCTransportConstants.CONNECTED_CLIENT :
        input=serverOfClient.getInputStream();
        output=serverOfClient.getOutputStream();
        break;
      case MJCTransportConstants.CONNECTED_SERVER :
        input=clientOfServer.getInputStream();
        output=clientOfServer.getOutputStream();
        break;
      default:
//        debugOutput("no streams initialized");
        break;
    }
  }

  /** Generates a checksum for a String.
      @param String: the String for which a checksum is needed
      @return String: the checksum
  */
  public String checkSum(String s) {
    long tempLong=0;
    String tempString;
    for (int i=0; i<s.length(); i++) {
      tempLong+=((int)s.charAt(i))+1;
    }
    tempString=new String((new Long(tempLong)).toString());
    if (tempString.length()==0) {
      tempString="##";
    }
    if (tempString.length()==1) {
      tempString=tempString.concat("#");
    }
    tempString=tempString.substring(tempString.length()-2,tempString.length());
    return tempString;
  }

  /** Writes debug information to the standard out.
      @param String s: the debug information
  */
  private void debugOutput(String s) {
    switch (status) {
      case MJCTransportConstants.CONNECTED_CLIENT :
        System.out.println("[TCP_C]"+s);
        break;
      case MJCTransportConstants.CONNECTED_SERVER :
        System.out.println("[TCP_S]"+s);
        break;
      default:
        System.out.println("[TCP_"+status+"]"+s);
        break;
    }
  }

  /** Initializes intern variables from the MJCConfiguration.
  */
  private void initConfiguration() {
    if (configuration==null) {
      readBuffer=new MJCStringBuffer(CONF_DEF_READ_BUFFER_SIZE);
      writeBuffer=new MJCStringBuffer(CONF_DEF_WRITE_BUFFER_SIZE);
    }
    else {
      readBuffer=new MJCStringBuffer(configuration.getInt(CONF_READ_BUFFER_SIZE,
        CONF_DEF_READ_BUFFER_SIZE));
      writeBuffer=new MJCStringBuffer(configuration.getInt(CONF_WRITE_BUFFER_SIZE,
        CONF_DEF_WRITE_BUFFER_SIZE));
    }
  }

}


