/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.util;

import com.monica.javacom.interfaces.*;
import java.util.*;
import java.io.*;

/** This class implements the MJConfiguration interface to provide customization possibility
    for MJC classes. This class reads customization information from a file called MJC.conf,
    which has to be located in the current directory.<br><br>
    The configuration file can have the following stucure:<br>
    <ul>
      <li>lines beginning with # are recognized as comments and therefore ignored
      <li>lines not beginning with # are recognized as configuration lines<br><br>
        configuration lines have the following structure:<br>
        <pre>keyword=value</pre>
        Values can be Strings, ints, and booleans. Booleans are represented with the
        strings "true" and "false". These values can be referenced int he get methods
        of the MJCConfiguration with the id matching the keyword.
    </ul>
*/
public abstract class MJCInputStreamConfiguration implements MJCConfiguration {

  // private variables
  private Properties prop;
  protected InputStream iS;
  private int status;

  // constants
  private static final int ST_NO=0;
  private static final int ST_YES=1;

  private static final String NAME="MJCInputStreamConfiguration";
  private static final String VERSION="0.1";

  protected void pre_init() {
    status=ST_NO;
  }

  protected void post_init() {
    try {
      prop=new Properties();
      prop.load(iS);
      status=ST_YES;
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  /** Sets the MJCConfiguration for this object.
      @param MJCConfiguration configuration: the MJCConfiguration
        object for the customization
  */
  public String getConfigurationName() {
    return NAME;
  }

  /** Returns the MJCConfiguration with which this object is customized.
      @return MJCConfiguration: the MJCConfiguration with which this
        object is customized
  */
  public String getConfigurationVersion() {
    return VERSION;
  }

  /** Retuns an int configuration value for the specified configuration id.
      @param String id: the required configuration id
      @return int: the configuration int matching the configuration id
      @exception Exception: if the id is invalid or the id doesn't represent an
        int value
  */
  public int getInt(String id) throws Exception {
    String temp;
    if (status==ST_YES) {
      temp=prop.getProperty(id);
      if (temp!=null) {
        return Integer.parseInt(temp);
      }
    }
    throw new Exception("Key not found in conf file - "+id);
  }

  /** Returns a boolean configuraton value for the specified configuration id.
      @param String id: the required configuration id
`      @return boolean: the configuration boolean matching the configuration id
      @exception Exception: if the id is invalid or the id doesn't represent a
        boolean value
  */
  public boolean getBoolean(String id) throws Exception {
    String temp;
    if (status==ST_YES) {
      temp=prop.getProperty(id);
      if (temp!=null) {
        if (temp.equals("true")) {
          return true;
        }
        if (temp.equals("false")) {
          return false;
        }
      }
    }
    throw new Exception("Key not found in conf file - "+id);
  }

  /** Returns a String configuraton value for the specified configuration id.
      @param String id: the required configuration id
      @return String: the configuration String matching the configuration id
      @exception Exception: if the id is invalid or the id doesn't represent a
        String value
  */
  public String getString(String id) throws Exception {
    String temp;
    if (status==ST_YES) {
      temp=prop.getProperty(id);
      if (temp!=null) {
        return temp;
      }
    }
    throw new Exception("Key not found in conf file - "+id);
  }

  /** Returns an int configuraton value for the specified configuration id.
      Should call the getInt(String) method of this interface and if there
      was an exception, returns the specified default value.
      @param String id: the required configuration id
      @param int def: the specified default value
      @return int: the configuration int matching the configuration id
        or if there was an exception then the default int
  */
  public int getInt(String id, int def) {
    try {
      return getInt(id);
    }
    catch (Exception e) {
      return def;
    }
  }

  /** Returns a boolean configuraton value for the specified configuration id.
      Should call the getBoolean(String) method of this interface and if there
      was an exception, returns the specified default value.
      @param String id: the required configuration id
      @param boolean def: the specified default value
      @return boolean: the configuration boolean matching the configuration id
        or if there was an exception then the default boolean
  */
  public boolean getBoolean(String id, boolean def) {
    try {
      return getBoolean(id);
    }
    catch (Exception e) {
      return def;
    }
  }

  /** Returns a String configuraton value for the specified configuration id.
      Should call the getString(String) method of this interface and if there
      was an exception, returns the specified default value.
      @param String id: the required configuration id
      @param boolean String: the specified default value
      @return boolean: the configuration String matching the configuration id
        or if there was an exception then the default String
  */
  public String getString(String id, String def) {
    try {
      return getString(id);
    }
    catch (Exception e) {
      return def;
    }
  }

  /** Useable if String possible values in the configuration source
      should be mapped into ints as configuration values.
      If there is a problem retrieving the int returned, the default
      value will be returned.
      @param String id: the required configuration id
      @param String[] possibilities: possible Strings acceptable in the
        configuration source
      @param int[] values: int values matching the possible Strings
      @param int def: the dafault int that should be returned in case of
        errors
      @return int: the int value matching the found possibility String of
        the def int
  */
  public int getIntFromStrings(String id, String[] possibilities, int[] values, int def) {
    String temp;
    int i;
    try {
      temp=getString(id);
      i=0;
      while (i<possibilities.length) {
        if (possibilities[i].equals(temp)) {
          return values[i];
        }
        i++;
      }
      return def;
    }
    catch (Exception e) {
      return def;
    }
  }

}