/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.util;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.util.*;
import java.util.*;

/** This class provides a buffer functionality. The buffer provides locking, adding of
    byte arrays, chars, char arrays, Strings, retrieving the buffer in synchronized
    way. The StringBuffer functionality means that the following types can be added, and
    a String can be received from the objects added, where the objects will be concatenated
    in the order of their addition to the buffer.
*/
public class MJCStringBuffer {

  private int bufferLength;
  private int maxBufferLength;
  private Vector buffer;
  private boolean locked;
  private MJCConverter converter;

  /** Constructs the buffer. The length of the contents of the buffer cannot be greater
      than <b>maxBufferLength</b>.
      @param int maxBufferLength: the maximal length of the contents of the buffer
  */
  public MJCStringBuffer(int maxBufferLength) {
    this.maxBufferLength=maxBufferLength;
    buffer=new Vector();
    locked=false;
    converter=new MJCConverter();
  }

  /** Locks the buffer. No content altering method can be invoked on a locked buffer.
  */
  public synchronized boolean lock() {
    if (!locked) {
      locked=true;
    }
    return locked;
  }

  /** Unlocks the buffer.
  */
  public synchronized void unlock() {
    locked=false;
  }

  /** Returns the maximal length of the contents of the buffer.
      @return int: the maximal length of the contents of the buffer
  */
  public synchronized int getMaxBufferLength() {
    return maxBufferLength;
  }

  /** Returns the actual length of the buffer.
      @return int: the actual length of the contents of the buffer
  */
  public synchronized int getBufferLength() {
    return bufferLength;
  }

  /** Adds a byte array to the buffer.
      @param byte[] b: the byte array to be added
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(byte[] b) {
    if (((bufferLength+b.length)>maxBufferLength) || (locked)) {
      return false;
    }
    else {
      buffer.addElement(converter.toString(b));
      bufferLength+=b.length;
      return true;
    }
  }

  /** Adds the first <b>l</b> bytes of the byte array to the buffer. Is
      equivalent to <b>add(b,0,l)</b>.
      @param byte b[]: the byte array to be added
      @param int l: the length of the byte array
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(byte[] b, int l) {
    if (((bufferLength+l)>maxBufferLength) || (locked)) {
      return false;
    }
    else {
      buffer.addElement(converter.toString(b,l));
      bufferLength+=l;
      return true;
    }
  }

  /** Adds a part of byte array specified with offset and length to the buffer.
      @param byte b[]: the byte array to be added
      @param int off: the beginning offset of the part (inclusive)
      @param int len: the length of the part
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(byte[] b, int off, int len) {
    if (((bufferLength+len)>maxBufferLength) || (locked)) {
      return false;
    }
    else {
      buffer.addElement(converter.toString(b,off,len));
      bufferLength+=len;
      return true;
    }
  }

  /** Adds a char array to the buffer.
      @param char[] c: the char array to be added
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(char[] c) {
    if (((bufferLength+c.length)>maxBufferLength) || (locked)) {
      return false;
    }
    else {
      buffer.addElement(new String(c));
      bufferLength+=c.length;
      return true;
    }
  }

  /** Adds a part of char array specified with offset and length to the buffer.
      @param char b[]: the char array to be added
      @param int off: the beginning offset of the part (inclusive)
      @param int len: the length of the part
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(char[] c, int off, int len) {
    if (((bufferLength+len)>maxBufferLength) || (locked)) {
      return false;
    }
    else {
      buffer.addElement(new String(c,off,len));
      bufferLength+=len;
      return true;
    }
  }

  /** Adds a single char to the buffer.
      @param char c: the char to be added
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(char c) {
    char[] ca=new char[1];
    ca[0]=c;
    return addToBuffer(ca);
  }

  /** Adds a String to the buffer.
      @param String s: the String to be added
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(String s) {
    if (((bufferLength+s.length())>maxBufferLength) || (locked)) {
      return false;
    }
    else {
      buffer.addElement(s);
      bufferLength+=s.length();
      return true;
    }
  }

  /** Adds a part of a String specified with offset and length to the buffer.
      @param String s: the String to be added
      @param int off: the beginning offset of the part (inclusive)
      @param int len: the length of the part
      @return boolean: wether the adding has been successful
  */
  public synchronized boolean addToBuffer(String s, int off, int len) {
    if (((bufferLength+len)>maxBufferLength) || (locked)) {
      return false;
    }
    else {
      buffer.addElement(s.substring(off,off+len));
      bufferLength+=len;
      return true;
    }
  }

  /** Returns the contents of the buffer. After this method is called, the contents
      of the buffer will be lost.
      @return String: the contents of the buffer
  */
  public synchronized String emptyBuffer() {
    if (!locked) {
      return usEmptyBuffer();
    }
    else {
      return null;
    }
  }

  /** Returns the contents of the buffer. After this method is called,
      the contents of the buffer will be lost.
      @return String: the contents of the buffer
  */
  private String usEmptyBuffer() {
    String temp=new String();
    for (int i=0; i<buffer.size(); i++) {
      temp=temp.concat((String)buffer.elementAt(i));
    }
    buffer=new Vector();
    bufferLength=0;
    return temp;
  }

  /** Returns the first <b>f</b> bytes of the buffer. If <b>f>getBufferLength()</b>
      then the whole buffer will be returned. The part of the buffer returned
      will be lost.
      @param int f: the length of the requested part of the buffer
      @return String: the part of the buffer
  */
  public synchronized String emptyBuffer(int f) {
    if (!locked) {
      if (f>bufferLength) {
        return usEmptyBuffer();
      }
      else {
        String tempRet=new String();
        String tempCurr, tempCurrB, tempCurrE;
	      boolean doLoop=true;
	      while (doLoop) {
 	        tempCurr=(String)buffer.elementAt(0);
	        buffer.removeElementAt(0);
	        if ((tempRet.length()+tempCurr.length())<f) {
	          tempRet=tempRet.concat(tempCurr);
	        }
	        else if ((tempRet.length()+tempCurr.length())==f) {
	          tempRet=tempRet.concat(tempCurr);
	          doLoop=false;
	        }
	        else {
	          tempCurrB=tempCurr.substring(0,f-tempRet.length());
	          tempCurrE=tempCurr.substring(f-tempRet.length(),
	          tempCurr.length());
	          tempRet=tempRet.concat(tempCurrB);
	          buffer.insertElementAt(tempCurrE,0);
	          doLoop=false;
	        }
	      }
	      return tempRet;
      }
    }
    else {
      return null;
    }
  }

  /** Returns the contents of the buffer. The contents of the bugger won't be lost.
      @return String: the contents of the buffer
  */
  public synchronized String getBuffer() {
    String temp=new String();
    for (int i=0; i<buffer.size(); i++) {
      temp=temp.concat((String)buffer.elementAt(i));
    }
    return temp;
  }

  /** Returns the status of the locking of the buffer.
      @return boolean: the status of the locking of the buffer. True - locked, false - unlocked.
  */
  public synchronized boolean lockStatus() {
    return locked;
  }

  /** Returns the number of bytes that can still be added to the buffer without
      rejection.
      @return int: the number of bytes that can still be added to the buffer
  */
  public synchronized int available() {
    return (maxBufferLength-bufferLength);
  }

  public synchronized boolean clearBuffer() {
    if (!locked) {
      buffer=new Vector();
      bufferLength=0;
      return true;
    }
    return false;
  }


}

