/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.virtualdisplay;

import com.monica.javacom.interfaces.*;
import java.awt.*;

/** This class provides the virtualDisplay functionality in the AWT framework.
    This class acts as a Canvas and can be added to other Containers.<br><br>
    Configureable values:
    <ul>
      <li><b>displayWidth</b>: default width of the screen in characters
      <li><b>displayHeight</b>: default width of the screen in characters
      <li><b>displayFillChar</b>: fill character for empty spaces
      <li><b>displayFontFace</b>: font face for the output
      <li><b>displayFontSize</b>: font size for the output
      <li><b>displayHorizontalMargin</b>: horizontal margin inside the window
      <li><b>displayVerticalMargin</b>: vertical margin inside the window
      <li><b>displayTabSize</b>: tab size in characters
      <li><b>displayNormalColor[R|G|B]</b>: normal color RGB values 0...255
      <li><b>displayBackgroundColor[R|G|B]</b>: backgroung color RGB values 0...255
      <li><b>displayInvertedColor[R|G|B]</b>: inverted color RGB values 0...255
      <li><b>displayInvertedBackgroundColor[R|G|B]</b>: inverted background RGB values 0...255
      <li><b>displayCursorColor[R|G|B]</b>: cursor color RGB values 0...255
    </ul>
*/
public class MJCAWTDisplay extends Canvas implements MJCVirtualDisplay {

  // ***************************************************************************************
  // Variables
  // ***************************************************************************************
  int posY;
  int posX;
  char[] content;
  boolean[] invertedContent;
  int sizeX;
  int sizeY;
  MJCConfiguration configuration;
  char fillchar;
  boolean[] changed;
  int verticalMargin, horizontalMargin, charWidth, lineHeight;
  Font displayFont;
  FontMetrics displayFontMetrics;
  boolean inverting;
  boolean moveCursor;
  Color normalColor;
  Color invertedColor;
  Color cursorColor;
  Color invertedBackgroundColor;
  Color normalBackgroundColor;
  int SRfirstLine, SRlastLine;

  int p_i, p_j, p_curr, c_i, c_j, sNL_i, sNL_j, sFC_i, sFC_j, sU_i, sU_j, aS_i, aCA_i, m_i, eEOL_i;

  // ***************************************************************************************
  // Constants
  // ***************************************************************************************

  private static final int CONF_CW_FIXCHAR=0;
  private static final int CONF_CW_MEASURE=1;
  private static final int CONF_CW_FIX=2;
  private static final int CONF_CW_FIXCHAROFFSET=3;

  // Configuration File
    private static final String CONF_HEIGHT="displayHeight";
    private static final String CONF_WIDTH="displayWidth";
    private static final String CONF_FILLCHAR="displayFillChar";
    private static final String CONF_FONTFACE="displayFontFace";
    private static final String CONF_FONTSIZE="displayFontSize";
    private static final String CONF_VERTICAL_MARGIN="displayHorizontalMargin";
    private static final String CONF_HORIZONTAL_MARGIN="displayVerticalMargin";
    private static final String CONF_NORMAL_COLOR_R="displayNormalColorR";
    private static final String CONF_NORMAL_COLOR_G="displayNormalColorG";
    private static final String CONF_NORMAL_COLOR_B="displayNormalColorB";
    private static final String CONF_INVERTED_COLOR_R="displayInvertedColorR";
    private static final String CONF_INVERTED_COLOR_G="displayInvertedColorG";
    private static final String CONF_INVERTED_COLOR_B="displayInvertedColorB";
    private static final String CONF_INVERTED_BACKGROUND_COLOR_R="displayInvertedBackgroundColorR";
    private static final String CONF_INVERTED_BACKGROUND_COLOR_G="displayInvertedBackgroundColorG";
    private static final String CONF_INVERTED_BACKGROUND_COLOR_B="displayInvertedBackgroundColorB";
    private static final String CONF_BACKGROUND_COLOR_R="displayBackgroundColorR";
    private static final String CONF_BACKGROUND_COLOR_G="displayBackgroundColorG";
    private static final String CONF_BACKGROUND_COLOR_B="displayBackgroundColorB";
    private static final String CONF_CURSOR_COLOR_R="displayCursorColorR";
    private static final String CONF_CURSOR_COLOR_G="displayCursorColorG";
    private static final String CONF_CURSOR_COLOR_B="displayCursorColorB";
    private static final String CONF_CHARWIDTH_TYPE="displayCharWidthType";
    private static final String CONF_CHARWIDTH_FIXCHAR="displayCharWidthFixChar";
    private static final String CONF_CHARWIDTH_FIX="displayCharWidthFix";
    private static final String CONF_CHARWIDTH_FIXCHAR_OFFSET="displayCharWidthFixCharOffset";

  // Default Values
    private static final int DEF_WIDTH=80; // default window width
    private static final int DEF_HEIGHT=25; //default window height
    private static final int DEF_FILLCHAR=32;
    private static final String DEF_FONTFACE="monospaced";
    private static final int DEF_FONTSIZE=10;
    private static final int DEF_VERTICAL_MARGIN=20;
    private static final int DEF_HORIZONTAL_MARGIN=10;
    private static final boolean DEF_HORIZONTAL_LINES=false;
    private static final boolean DEF_VERTICAL_LINES=false;
    private static final int DEF_NORMAL_COLOR_R=0;
    private static final int DEF_NORMAL_COLOR_G=0;
    private static final int DEF_NORMAL_COLOR_B=0;
    private static final int DEF_INVERTED_COLOR_R=255;
    private static final int DEF_INVERTED_COLOR_G=255;
    private static final int DEF_INVERTED_COLOR_B=255;
    private static final int DEF_INVERTED_BACKGROUND_COLOR_R=0;
    private static final int DEF_INVERTED_BACKGROUND_COLOR_G=0;
    private static final int DEF_INVERTED_BACKGROUND_COLOR_B=0;
    private static final int DEF_BACKGROUND_COLOR_R=255;
    private static final int DEF_BACKGROUND_COLOR_G=255;
    private static final int DEF_BACKGROUND_COLOR_B=255;
    private static final int DEF_CURSOR_COLOR_R=255;
    private static final int DEF_CURSOR_COLOR_G=0;
    private static final int DEF_CURSOR_COLOR_B=0;
    private static final int DEF_CHARWIDTH_TYPE=CONF_CW_FIXCHAR;
    private static final int DEF_CHARWIDTH_FIX=10;
    private static final int DEF_CHARWIDTH_FIXCHAR=95;
    private static final int DEF_CHARWIDTH_FIXCHAR_OFFSET=10;

  // arrays
    private static final String[] CONF_CW_POSSIBILITIES={"FIXCHAR", "MEASURE", "FIX", "FIXCHAROFFSET"};
    private static final int[] CONF_CW_VALUES={CONF_CW_FIXCHAR, CONF_CW_MEASURE, CONF_CW_FIX,
      CONF_CW_FIXCHAROFFSET};

    private static final String NAME="MJCAWTDisplay";
    private static final String VERSION="0.1";

  // ***************************************************************************************
  // Initialization
  // ***************************************************************************************

  /** Default constructor. The MJCConfiguration for customization has to be provided.
      @param MJCConfiguration configuration: the customization object
  */
  public MJCAWTDisplay(MJCConfiguration configuration) {
    this.configuration=configuration;

    rereadConf();
  }

  private void rereadConf() {
    sizeX=configuration.getInt(CONF_WIDTH, DEF_WIDTH);
    sizeY=configuration.getInt(CONF_HEIGHT, DEF_HEIGHT);

    fillchar=(char)configuration.getInt(CONF_FILLCHAR, DEF_FILLCHAR);

    horizontalMargin=configuration.getInt(CONF_HORIZONTAL_MARGIN, DEF_HORIZONTAL_MARGIN);
    verticalMargin=configuration.getInt(CONF_VERTICAL_MARGIN, DEF_VERTICAL_MARGIN);

    displayFont=new Font(configuration.getString(CONF_FONTFACE, DEF_FONTFACE), Font.PLAIN,
      configuration.getInt(CONF_FONTSIZE, DEF_FONTSIZE));
    displayFontMetrics=getFontMetrics(displayFont);

    // charWidth

    charWidth=0;
    switch (configuration.getIntFromStrings(CONF_CHARWIDTH_TYPE, CONF_CW_POSSIBILITIES,
      CONF_CW_VALUES, DEF_CHARWIDTH_TYPE)) {
      case CONF_CW_FIX:
        charWidth=configuration.getInt(CONF_CHARWIDTH_FIX, DEF_CHARWIDTH_FIX);
        break;
      case CONF_CW_MEASURE:
        for (int i=31; i<256; i++) {
          if ((displayFontMetrics.charWidth(i)+1)>charWidth) {
            charWidth=displayFontMetrics.charWidth(i)+1;
          }
        }
        break;
      case CONF_CW_FIXCHAR:
        charWidth=displayFontMetrics.charWidth(configuration.getInt(CONF_CHARWIDTH_FIXCHAR,
          DEF_CHARWIDTH_FIXCHAR))+1;
        break;
      case CONF_CW_FIXCHAROFFSET:
        charWidth=displayFontMetrics.charWidth(configuration.getInt(CONF_CHARWIDTH_FIXCHAR,
          DEF_CHARWIDTH_FIXCHAR))+configuration.getInt(CONF_CHARWIDTH_FIXCHAR_OFFSET,
          DEF_CHARWIDTH_FIXCHAR_OFFSET);
        break;
    }

    lineHeight=displayFontMetrics.getHeight();

    setFont(displayFont);
    initialize();
    setSize(new Dimension(sizeX*charWidth+2*horizontalMargin, sizeY*lineHeight+2*verticalMargin));

    normalColor=new Color(
      configuration.getInt(CONF_NORMAL_COLOR_R, DEF_NORMAL_COLOR_R),
      configuration.getInt(CONF_NORMAL_COLOR_G, DEF_NORMAL_COLOR_G),
      configuration.getInt(CONF_NORMAL_COLOR_B, DEF_NORMAL_COLOR_B));
    invertedColor=new Color(
      configuration.getInt(CONF_INVERTED_COLOR_R, DEF_INVERTED_COLOR_R),
      configuration.getInt(CONF_INVERTED_COLOR_G, DEF_INVERTED_COLOR_G),
      configuration.getInt(CONF_INVERTED_COLOR_B, DEF_INVERTED_COLOR_B));
    invertedBackgroundColor=new Color(
      configuration.getInt(CONF_INVERTED_BACKGROUND_COLOR_R, DEF_INVERTED_BACKGROUND_COLOR_R),
      configuration.getInt(CONF_INVERTED_BACKGROUND_COLOR_G, DEF_INVERTED_BACKGROUND_COLOR_G),
      configuration.getInt(CONF_INVERTED_BACKGROUND_COLOR_B, DEF_INVERTED_BACKGROUND_COLOR_B));
    normalBackgroundColor=new Color(
      configuration.getInt(CONF_BACKGROUND_COLOR_R, DEF_BACKGROUND_COLOR_R),
      configuration.getInt(CONF_BACKGROUND_COLOR_G, DEF_BACKGROUND_COLOR_G),
      configuration.getInt(CONF_BACKGROUND_COLOR_B, DEF_BACKGROUND_COLOR_B));
    cursorColor=new Color(
      configuration.getInt(CONF_CURSOR_COLOR_R, DEF_CURSOR_COLOR_R),
      configuration.getInt(CONF_CURSOR_COLOR_G, DEF_CURSOR_COLOR_G),
      configuration.getInt(CONF_CURSOR_COLOR_B, DEF_CURSOR_COLOR_B));
    setBackground(normalBackgroundColor);
  }

  /** Sets the MJCConfiguration for this object.
      @param MJCConfiguration configuration: the MJCConfiguration
        object for the customization
  */
  public void setConfiguration(MJCConfiguration configuration) {
    this.configuration=configuration;
    rereadConf();
  }

  /** Returns the MJCConfiguration with which this object is customized.
      @return MJCConfiguration: the MJCConfiguration with which this
        object is customized
  */
  public MJCConfiguration getConfiguration() {
    return configuration;
  }

  /** Returns the MJC name of the MJCVirtualDisplay.
      @return String: the MJC name of the MJCVirtualDisplay
  */
  public String getVirtualDisplayName() {
    return NAME;
  }

  /** Returns the MJC version of the MJCVirtualDisplay.
      @return String: the MJC name of the MJCVirtualDisplay
  */
  public String getVirtualDisplayVersion() {
    return VERSION;
  }

  /** Initializes intern variables with their default values.
  */
  private void initialize() {
    posX=0;
    posY=0;
    inverting=false;
    content=new char[sizeX*sizeY+1];
    invertedContent=new boolean[sizeX*sizeY+1];
    changed=new boolean[sizeX*sizeY+1];
    moveCursor=true;
    SRfirstLine=0;
    SRlastLine=sizeY-1;
    clear();
  }

  /** Repaints the contents of the Canvas.
      @param Grpahics g: the Graphics object with which the rapinting works
  */
  public void paint(Graphics g) {
    redraw(g, true);
  }

  /** Redraws the character screen on the Canvas.
      @param Grpahics g: the Graphics object with which the radrawing works
  */
  private void redraw(Graphics g, boolean painting) {
    debugOutput("paint begin");

    p_curr=0;

//    g.setColor(new Color(0,0,0));

    int lP, tP;

    for (p_i=0; p_i<sizeY; p_i++) {
      for (p_j=0; p_j<sizeX; p_j++) {

        if ((changed[p_curr]) || (painting)) {
          changed[p_curr]=false;

          lP=horizontalMargin+p_j*charWidth;
          tP=verticalMargin+p_i*lineHeight;

          if (invertedContent[p_curr]) { // inverted
            g.setColor(invertedBackgroundColor);
            g.fillRect(lP, tP, charWidth, lineHeight);
            g.setColor(invertedColor);
          }
          else { // normal
            g.setColor(normalBackgroundColor);
            g.fillRect(lP, tP, charWidth, lineHeight);
            g.setColor(normalColor);
          }
          if (content[p_curr]!=fillchar) {
            g.drawString(""+content[p_curr], lP, tP+lineHeight-2);
          }

        }

        p_curr++;
      }
//      g.setColor(new Color(100,255,255));
    }

    if (posX!=sizeX) {
      lP=horizontalMargin+posX*charWidth;
      tP=verticalMargin+posY*lineHeight;
      g.setColor(cursorColor);
      g.drawRect(lP, tP, charWidth-1, lineHeight-1);
    }

    debugOutput("paint end");
  }

  // ***************************************************************************************
  // Size Management
  // ***************************************************************************************

  /** Sets the screen size in characters. After resize the screen will
      be cleared.
      @param int sizeX: the width
      @param int sizeY: the height
  */
  public void setCSize(int sizeX, int sizeY) {
    this.sizeX=sizeX;
    this.sizeY=sizeY;
    initialize();
  }

  /** Sets the screen width in characters. After resize the screen will
      be cleared.
      @param int sizeX: the width
  */
  public void setCSizeX(int sizeX) {
    this.sizeX=sizeX;
    initialize();
  }

  /** Sets the screen height in characters. After resize the screen will
      be cleared.
      @param int sizeY: the height
  */
  public void setCSizeY(int sizeY) {
    this.sizeY=sizeY;
    initialize();
  }

  /** Returns the screen width in characters.
      @return int: the width
  */
  public int getCSizeX() {
    return sizeX;
  }

  /** Returns the screen height in characters.
      @return int: the height
  */
  public int getCSizeY() {
    return sizeY;
  }

  // ***************************************************************************************
  // Global Display Management
  // ***************************************************************************************

  /** Clears the contents of the screen. The position of the cursor will
      not be changed.
  */
  public void clear() {
    for (c_i=0; c_i<sizeY; c_i++) {
      for (c_j=0; c_j<sizeX; c_j++) {
        content[c_i*sizeX+c_j]=fillchar;
        invertedContent[c_i*sizeX+c_j]=false;
	      changed[c_i*sizeX+c_j]=true;
      }
    }
  }

  // ***************************************************************************************
  // Current Position Management
  // ***************************************************************************************

  /** Marks the current position of the cursor to changed for the next redraw.
  */
  private void setCurrChanged() {
//    debugOutput("sCC");
    changed[posY*sizeX+posX]=true;
  }

  /** Sets the position of the cursor. Indexing begins with 0.
      @param int x: the column
      @param int y: the row
  */
  public void setCPos(int x, int y) {
    setCurrChanged();
    posX=x;
    posY=y;
    setCurrChanged();
  }

  /** Sets the column of the cursor. Indexing begins with 0.
      @param int x: the column
  */
  public void setCPosX(int x) {
    setCurrChanged();
    posX=x;
    setCurrChanged();
  }

  /** Sets the row of the cursors. Indexing begins with 0.
      @param int y: the row
  */
  public void setCPosY(int y) {
    setCurrChanged();
    posY=y;
    setCurrChanged();
  }

  /** Returns the current column of the cursor. Indexing begins with 0.
      @return int: the current column
  */
  public int getCPosX() {
    return posX;
  }

  /** Returns the current row of the cursor. Indexing begins with 0.
      @return int: the current row
  */
  public int getCPosY() {
    return posY;
  }

  // ***************************************************************************************
  // Typing Management
  // ***************************************************************************************

  /** Adds a character to the current cursor position. The cursor will
      be adwanced once. The cursor can be positioned one characted right
      off the screen. If there is an add character at that position,
      the cursor will go one line down and to the beginning of the line,
      and the character will be added there. If there are no lines to
      go down, the screen will be scrolled down. The removed line at the top
      will be lost. The cursor cannot leave the scrolling region with this
      method. Characters will be added with the current inverting status.
      @param char c: the characted to be added
  */
  public void addC(char c) {
    setCurrChanged();

    if (posX==sizeX) {
      posX=0;
      posY++;
      if (posY>SRlastLine) {
        posY--;
        scrollNewLine();
      }
    }

//    debugOutput("char added : posX="+posX+", posY="+posY+" char="+(int)c+", inverting="+inverting);

    content[posY*sizeX+posX]=c;
    invertedContent[posY*sizeX+posX]=inverting;

    if (moveCursor) {
      ++posX;
    }

    setCurrChanged();
  }

  /** Creates a new empty row at the end of the scrolling region.
  */
  private void scrollNewLine() {
    for (sNL_i=SRfirstLine; sNL_i<SRlastLine; sNL_i++) {
      for (sNL_j=0; sNL_j<sizeX; sNL_j++) {
        content[sNL_i*sizeX+sNL_j]=content[(sNL_i+1)*sizeX+sNL_j];
 	      invertedContent[sNL_i*sizeX+sNL_j]=invertedContent[(sNL_i+1)*sizeX+sNL_j];
      }
    }
    for (sNL_i=0; sNL_i<sizeX; sNL_i++) {
      content[SRlastLine*sizeX+sNL_i]=fillchar;
      invertedContent[SRlastLine*sizeX+sNL_i]=false;
    }
    setFullChanged();
  }

  /** Marks the whole screen changed for the next redraw.
  */
  private void setFullChanged() {
    for (int sFC_i=0; sFC_i<sizeY; sFC_i++) {
      for (int sFC_j=0; sFC_j<sizeX; sFC_j++) {
        changed[sizeX*sFC_i+sFC_j]=true;
      }
    }
  }

  /** Moves all rows int the scrolling region one row up. The last
      row in the scrolling region will get empty. Cursor position
      will not be altered.
  */
  public void scrollDown() {
    newLine();
  }

  /** Moves all rows in the scolling region one row down. The first row
      in the scrolling region will get empty. Cursor position will
      not be altered.
  */
  public void scrollUp() {
    for (sU_i=SRlastLine; sU_i>SRfirstLine; sU_i--) {
      for (sU_j=0; sU_j<sizeX; sU_j++) {
        content[sU_i*sizeX+sU_j]=content[(sU_i-1)*sizeX+sU_j];
	      invertedContent[sU_i*sizeX+sU_j]=invertedContent[(sU_i-1)*sizeX+sU_j];
      }
    }
    for (sU_i=0; sU_i<sizeX; sU_i++) {
      content[SRfirstLine*sizeX+sU_i]=fillchar;
      invertedContent[SRfirstLine*sizeX+sU_i]=false;
    }
    setFullChanged();
  }

  /** Cursors goes one row down. If the current cursor position is at the bottom
      of the screen, the screen will be scrolled down. The row scrolled
      out will be lost. Horizontal position of the cursor will not be altered.
  */
  public void newLine() {
    setCurrChanged();
    posY++;
    if (posY>SRlastLine) {
      posY--;
      scrollNewLine();
    }
    setCurrChanged();
  }

  /** Adds a String to the current cursor position. This method should do
      <b>s.length()</b> times the functionality of
      <b>addC(s.charAt(i))</b>.
      @param String s: the String to be added
  */
  public void addS(String s) {
    for (aS_i=0; aS_i<s.length(); aS_i++) {
      addC(s.charAt(aS_i));
    }
  }

  /** Adds a char array to the current cursor position. This method
      should do <b>ca.length</b> times the functionality of
      <b>addC(ca[i])</b>.
      @param char[] ca: the char array to be added
  */
  public void addCA(char[] c) {
    for (aCA_i=0; aCA_i<c.length; aCA_i++) {
      addC(c[aCA_i]);
    }
  }

  /** Refreshes the screen contents for the user.
  */
  public void refresh() {
    Graphics g=getGraphics();
    if (g!=null) {
      redraw(g, false);
    }
  }

  /** Cursor goes one column to the left. If the current cursor position is the
      first column then the cursor will go up one row (if possible) and to
      the last column. There won't be any scrolling. The cursor cannot leave
      the scrolling region with this method.
  */
  public boolean goLeft() {
    setCurrChanged();
    if (posX>0) {
      posX--;
      setCurrChanged();
      return true;
    }
    else {
      if (posY>SRfirstLine) {
        posY--;
	      posX=sizeX-1;
	      setCurrChanged();
	      return true;
      }
    }
    return false;
  }

  /** Cursor goes <>num</b> times to the left. This method should do
      <b>num</b> times the functionality of <b>goLeft()</b>.
      @param int num: the number of times to go left
  */
  public int goLeft(int num) {
    for (m_i=0; m_i<num; m_i++) {
      if (!goLeft()) {
        return m_i;
      }
    }
    return num;
  }

  /** Cursor goes one column to the right. If the current cursor position is
      the last column then the cursor will go down one row (if possible) and
      to the first column. There won't be any scrolling. The cursor cannot
      leave the scrolling region wit this method.
  */
  public boolean goRight() {
    setCurrChanged();
    if (posX<(sizeX-1)) {
      posX++;
      setCurrChanged();
      return true;
    }
    else {
      if (posY<SRlastLine) {
        posX=0;
	      posY++;
	      setCurrChanged();
	      return true;
      }
    }
    return false;
  }

  /** Cursor goes <b>num</b> times to the right. This method should do
      <b>num</b> times the functionality of <b>goRight()</b>.
      @param int num: the number of times to go right
  */
  public int goRight(int num){
    for (m_i=0; m_i<num; m_i++) {
      if (!goRight()) {
        return m_i;
      }
    }
    return num;
  }

  /** Cursor goes one row up (if possible). There won't be any scrolling.
      Horizontal position of the cursor will not be altered. Cursor
      cannot leave the scrolling region with this method.
  */
  public boolean goUp() {
    setCurrChanged();
    if (posY>SRfirstLine) {
      posY--;
      setCurrChanged();
      return true;
    }
    return false;
  }

  /** Cursor goes <b>num</b> times up. This method should do
      <b>num</b> times the functionality of <b>goUp()</b>.
      @param int num: the number of times to go up
  */
  public int goUp(int num) {
    for (m_i=0; m_i<num; m_i++) {
      if (!goUp()) {
        return m_i;
      }
    }
    return num;
  }

  /** Cursor goes one row down (if possible). There won't be any scrolling.
      Horizontal position of the cursor will not be altered. Cursor
      cannot leave the scrolling region with this method.
  */
  public boolean goDown() {
    setCurrChanged();
    if (posY<SRlastLine) {
      ++posY;
      setCurrChanged();
      return true;
    }
    return false;
  }

  /** Cursor goes <b>num</b> times down. This method should do
      <b>num</b> times the functionality of <b>goDown()</b>.
      @param int num: the number of times to go down.
  */
  public int goDown(int num) {
    for (m_i=0; m_i<num; m_i++) {
      if (!goDown()) {
        return m_i;
      }
    }
    return num;
  }

  /** Sets the inverting status. True means inverted, false normal.
      @param boolan status: the inverting status
  */
  public void setInvert(boolean status) {
    inverting=status;
    setCurrChanged();
  }

  /** Writes debug information to the standard output.
      @param String s: the debug information
  */
  private void debugOutput(String s) {
//    System.out.println("[VD]"+s);
  }

  /** Dumps the contents of the screen to the standard output.
  */
  public void dumpContents() {
    debugOutput("dumping content");
    debugOutput("posX="+posX+", posY="+posY);
    for (int i=0; i<sizeX*sizeY; i++) {
      debugOutput("dC("+i+")="+content[i]+", i="+invertedContent[i]);
    }
  }

  /** Erases the contents of the current row from the horizontal cursor
      position (inclusive) to the last column (inclusive).
  */
  public void eraseEOL() {
    for (int eEOL_i=posX; eEOL_i<sizeX; eEOL_i++) {
      content[posY*sizeX+eEOL_i]=fillchar;
      invertedContent[posY*sizeX+eEOL_i]=false;
      changed[posY*sizeX+eEOL_i]=true;
    }
  }

  /** Should addC/addCA/addS advance the cursor while adding characters.
      True means yes, false no. If the cursor is positioned outside the
      visible screen area and an addC/addCA/addS method is called,
      the cursor will in either case go one row down and to the first
      column (and a scrolling will be performed if required). By default
      addC/addCA/addS will advance the cursor.
      @param boolean status: should the addC/addCA/addS methods advance
        the cursor while adding characters.
  */
  public void moveCursor(boolean status) {
    moveCursor=status;
  }

  /** Sets the scrolling region for the screen. Scrolling region is
      given by the first and last line (both includive, indexing begins
      with 0). If the cursor is inside the scrolling are, it
      cannot leave it with addC/addCA/addS/goLeft/goRight/goUp/goDown/
      newLine/insertLine/insertChar/deleteLine/deleteChar. By
      default the scrolling region is the whole screen.
      @param int firstLine: the first line of the scrolling region
      @param int lastLine: the last line of the scrolling region
  */
  public void setScrollingRegion(int firstLine, int lastLine) {
    SRfirstLine=firstLine;
    SRlastLine=lastLine;
  }

  /** Inserts <b>lineNum</b> rows at the current cursor position.
      The current row and the rows below the cursor will be moved
      <b>lineNum</b> rows down. The last row in the scrolling
      region will get lost during the movement. <b>lineNum</b> empty rows
      will be created beginning with the current row. Cursor will move
      to the first column of the current row.
      @param int lineNum: the number of rows to be inserted
  */
  public void insertLine(int lineNum) {
//    debugOutput("posX="+posX+" posY="+posY);
    setCPosX(0);
    for (int i=0; i<lineNum; i++) {
      for (int j=SRlastLine; j>posY; j--) {
        for (int k=0; k<sizeX; k++) {
          content[j*sizeX+k]=content[(j-1)*sizeX+k];
          invertedContent[j*sizeX+k]=invertedContent[(j-1)*sizeX+k];
        }
      }
      for (int l=0; l<sizeX; l++) {
        content[posY*sizeX+l]=fillchar;
        invertedContent[posY+sizeX+l]=inverting;
      }
    }
    setFullChanged();
  }

  /** Deletes <b>lineNum</b> rows at the current cursor position.
      The rows below the deleted rows will be moved upwards. At the bottom
      of the scrolling region new empty lines will be added. Cursor
      will move to the first column of the current row.
      @param int lineNum: the number of rows to be deleted
  */
  public void deleteLine(int lineNum) {
//    debugOutput("posX="+posX+" posY="+posY+" lineNum="+lineNum+" SRlastLine="+SRlastLine);
    setCPosX(0);

    for (int i=0; i<lineNum; i++) {
      for (int j=posY; j<SRlastLine; j++) {
        for (int k=0; k<sizeX; k++) {
          content[j*sizeX+k]=content[(j+1)*sizeX+k];
          invertedContent[j*sizeX+k]=invertedContent[(j+1)*sizeX+k];
        }
      }
      for (int l=0; l<sizeX; l++) {
        content[SRlastLine*sizeX+l]=fillchar;
        invertedContent[SRlastLine+sizeX+l]=inverting;
      }
    }

    setFullChanged();
  }

  /** Deletes a character at the cursor position. Characters after the
      cursor position in the current row will be moved one position to
      the left. The last character in the current row will be empty.
  */
  public void deleteChar() {
    setCurrChanged();
    for (int i=posX; i<(sizeX-1); i++) {
      content[posY*sizeX+i]=content[posY*sizeX+i+1];
      invertedContent[posY*sizeX+i]=invertedContent[posY*sizeX+i+1];
      changed[posY*sizeX+i]=true;
    }
    content[(posY+1)*sizeX-1]=fillchar;
    invertedContent[(posY+1)*sizeX-1]=false;
    changed[(posY+1)*sizeX-1]=true;
  }

  /** Deletes <b>charNum</b> characters at the cursor position.
      This method should do <b>charNum</b> times the functionality
      of <b>deleteChar()</b>.
      @param int charNum: number of characters to be deleted
  */
  public void deleteChar(int charNum) {
    for (int i=0; i<charNum; i++) {
      deleteChar();
    }
  }

  /** Inserts an empty character at the cursor position. Characters after
      the cursor position in the current row will be moved one position
      to the right. The last character in the current row will be lost.
  */
  public void insertChar() {
    for (int i=sizeX-1; i>posX; i--) {
      content[posY*sizeX+i]=content[posY*sizeX+i-1];
      invertedContent[posY*sizeX+i]=invertedContent[posY*sizeX+i-1];
      changed[posY*sizeX+i]=true;
    }
    content[posY*sizeX+posX]=fillchar;
    invertedContent[posY*sizeX+posX]=inverting;
    changed[posY*sizeX+posX]=true;
  }

  /** Inserts <b>charNum</b> empty characters at the cursor position.
      This method should do <b>charNum</b> times the functionality of
      <b>insertChar()</b>.
      @param int charNum: number of characters to be inserted
  */
  public void insertChar(int charNum) {
    for (int i=0; i<charNum; i++) {
      insertChar();
    }
  }
}